from typing import Any, Iterator

import pandas as pd

from qubx import logger
from qubx.core.basics import DataType, Instrument, Timestamped
from qubx.core.exceptions import SimulationError
from qubx.data.composite import IteratedDataStreamsSlicer
from qubx.data.readers import (
    AsDict,
    AsOrderBook,
    AsQuotes,
    AsTrades,
    DataReader,
    DataTransformer,
    RestoredBarsFromOHLC,
    RestoreQuotesFromOHLC,
    RestoreTradesFromOHLC,
)


class DataFetcher:
    _fetcher_id: str
    _reader: DataReader
    _requested_data_type: str
    _producing_data_type: str
    _params: dict[str, object]
    _specs: list[str]

    _transformer: DataTransformer
    _timeframe: str | None = None
    _warmup_period: pd.Timedelta | None = None
    _chunksize: int = 5000

    def __init__(
        self,
        fetcher_id: str,
        reader: DataReader,
        subtype: str,
        params: dict[str, Any],
        warmup_period: pd.Timedelta | None = None,
        chunksize: int = 5000,
        open_close_time_indent_secs=1.0,  # open/close shift may depends on simulation
    ) -> None:
        self._fetcher_id = fetcher_id
        self._params = params
        self._reader = reader

        match subtype:
            case DataType.OHLC_QUOTES:
                # - requested restore quotes from OHLC
                self._transformer = RestoreQuotesFromOHLC(open_close_time_shift_secs=open_close_time_indent_secs)
                self._requested_data_type = "ohlc"
                self._producing_data_type = "quote"
                if "timeframe" in params:
                    self._timeframe = params.get("timeframe", "1Min")

            case DataType.OHLC_TRADES:
                # - requested restore trades from OHLC
                self._transformer = RestoreTradesFromOHLC(open_close_time_shift_secs=open_close_time_indent_secs)
                self._requested_data_type = "ohlc"
                self._producing_data_type = "trade"
                if "timeframe" in params:
                    self._timeframe = params.get("timeframe", "1Min")

            case DataType.OHLC:
                # - requested restore bars from OHLC
                self._transformer = RestoredBarsFromOHLC(open_close_time_shift_secs=open_close_time_indent_secs)
                self._requested_data_type = "ohlc"
                self._producing_data_type = "ohlc"
                if "timeframe" in params:
                    self._timeframe = params.get("timeframe", "1Min")

            case DataType.TRADE:
                self._requested_data_type = "trade"
                self._producing_data_type = "trade"
                self._transformer = AsTrades()

            case DataType.QUOTE:
                # self._requested_data_type = "orderbook"
                self._requested_data_type = "quote"
                self._producing_data_type = "quote"  # ???
                self._transformer = AsQuotes()

            case DataType.ORDERBOOK:
                self._requested_data_type = "orderbook"
                self._producing_data_type = "orderbook"
                self._transformer = AsOrderBook()

            case _:
                self._requested_data_type = subtype
                self._producing_data_type = subtype
                self._transformer = AsDict()

        self._warmup_period = warmup_period
        self._warmed = {}
        self._specs = []
        self._chunksize = chunksize

    @staticmethod
    def _make_request_id(instrument: Instrument) -> str:
        return f"{instrument.exchange}:{instrument.symbol}"

    def attach_instrument(self, instrument: Instrument) -> str:
        _data_id = self._make_request_id(instrument)

        if _data_id not in self._specs:
            self._specs.append(_data_id)
            self._warmed[_data_id] = False

        return self._fetcher_id + "." + _data_id

    def remove_instrument(self, instrument: Instrument) -> str:
        _data_id = self._make_request_id(instrument)

        if _data_id in self._specs:
            self._specs.remove(_data_id)
            del self._warmed[_data_id]

        return self._fetcher_id + "." + _data_id

    def has_instrument(self, instrument: Instrument) -> bool:
        return self._make_request_id(instrument) in self._specs

    def get_instruments_indices(self) -> list[str]:
        return [self._fetcher_id + "." + i for i in self._specs]

    def get_instrument_index(self, instrument: Instrument) -> str:
        return self._fetcher_id + "." + self._make_request_id(instrument)

    def load(
        self, start: str | pd.Timestamp, end: str | pd.Timestamp, to_load: list[Instrument] | None
    ) -> dict[str, Iterator]:
        """
        Loads data for specified instruments within a given time range.

        Parameters:
            - start (str | pd.Timestamp): The start time for data loading, can be a string or a pandas Timestamp.
            - end (str | pd.Timestamp): The end time for data loading, can be a string or a pandas Timestamp.
            - to_load (list[Instrument] | None): A list of instruments to load data for. If None, data for all subscribed instruments is loaded.

        Returns:
            - dict[str, Iterator]: A dictionary where keys are instrument identifiers and values are iterators over the loaded data.
        """
        _requests = self._specs if not to_load else set(self._make_request_id(i) for i in to_load)
        _r_iters = {}

        for _r in _requests:  # - TODO: replace this loop with multi-instrument request after DataReader refactoring
            if _r in self._specs:
                _start = pd.Timestamp(start)
                if self._warmup_period and not self._warmed.get(_r):
                    _start -= self._warmup_period
                    self._warmed[_r] = True

                _args = dict(
                    data_id=_r,
                    start=_start,
                    stop=end,
                    transform=self._transformer,
                    data_type=self._requested_data_type,
                    chunksize=self._chunksize,
                )

                if self._timeframe:
                    _args["timeframe"] = self._timeframe

                # get arguments from self._reader.read
                _reader_args = set(self._reader.read.__code__.co_varnames[: self._reader.read.__code__.co_argcount])
                # match _args with self._params
                _params = {k: v for k, v in self._params.items() if k in _reader_args}
                _args = {**_args, **_params}

                try:
                    _r_iters[self._fetcher_id + "." + _r] = self._reader.read(**_args)  # type: ignore
                except Exception as e:
                    logger.error(f">>> (DataFetcher::load) - failed to load <g>'{self._fetcher_id}'</g> data: {e}")
            else:
                raise IndexError(
                    f"Instrument {_r} is not subscribed for this data {self._requested_data_type} in {self._fetcher_id} !"
                )

        return _r_iters

    def __repr__(self) -> str:
        return f"{self._requested_data_type}({self._params}) (-{self._warmup_period if self._warmup_period else '--'}) [{','.join(self._specs)}] :-> {self._transformer.__class__.__name__}"


class IterableSimulationData(Iterator):
    """
    This class is a crucial component for backtesting system.
    It provides a flexible and efficient way to simulate market data feeds for strategy testing.

    Key Features:
        - Supports multiple data types (OHLC, trades, quotes) and instruments.
        - Allows for dynamic addition and removal of instruments during simulation.
        - Handles warmup periods for data preloading.
        - Manages historical and current data distinction during iteration.
        - Utilizes a data slicer (IteratedDataStreamsSlicer) to merge and order data from multiple sources.

    TODO:
        1. think how to provide initial "market quote" for each instrument
        2. optimization for historical data (return bunch of history instead of each bar in next(...))
    """

    _readers: dict[str, DataReader]
    _subtyped_fetchers: dict[str, DataFetcher]
    _warmups: dict[str, pd.Timedelta]
    _instruments: dict[str, tuple[Instrument, DataFetcher, DataType]]
    _open_close_time_indent_secs: int | float

    _slicer_ctrl: IteratedDataStreamsSlicer | None = None
    _slicing_iterator: Iterator | None = None
    _start: pd.Timestamp | None = None
    _stop: pd.Timestamp | None = None
    _current_time: int | None = None

    def __init__(
        self,
        readers: dict[str, DataReader],
        open_close_time_indent_secs=1,  # open/close ticks shift
    ):
        self._readers = dict(readers)
        self._instruments = {}
        self._subtyped_fetchers = {}
        self._warmups = {}
        self._open_close_time_indent_secs = open_close_time_indent_secs

    def set_typed_reader(self, data_type: str, reader: DataReader):
        self._readers[data_type] = reader
        if _fetcher := self._subtyped_fetchers.get(data_type):
            _fetcher._reader = reader

    def set_warmup_period(self, subscription: str, warmup_period: str | None = None):
        if warmup_period:
            _access_key, _, _ = self._parse_subscription_spec(subscription)
            self._warmups[_access_key] = pd.Timedelta(warmup_period)

    def _parse_subscription_spec(self, subscription: str) -> tuple[str, str, dict[str, object]]:
        _subtype, _params = DataType.from_str(subscription)
        match _subtype:
            case DataType.OHLC | DataType.OHLC_QUOTES:
                _timeframe = _params.get("timeframe", "1Min")
                _access_key = f"{_subtype}.{_timeframe}"
            case DataType.TRADE | DataType.QUOTE | DataType.ORDERBOOK:
                _access_key = f"{_subtype}"
            case _:
                # - any arbitrary data type is passed as is
                _params = {}
                _subtype = subscription
                _access_key = f"{_subtype}"
        return _access_key, _subtype, _params

    def add_instruments_for_subscription(self, subscription: str, instruments: list[Instrument] | Instrument):
        instruments = instruments if isinstance(instruments, list) else [instruments]
        _subt_key, _data_type, _params = self._parse_subscription_spec(subscription)

        fetcher = self._subtyped_fetchers.get(_subt_key)
        if not fetcher:
            _reader = self._readers.get(_data_type)

            if _reader is None:
                raise SimulationError(f"No reader configured for data type: {_data_type}")

            self._subtyped_fetchers[_subt_key] = (
                fetcher := DataFetcher(
                    _subt_key,
                    _reader,
                    _data_type,
                    _params,
                    warmup_period=self._warmups.get(_subt_key),
                    open_close_time_indent_secs=self._open_close_time_indent_secs,
                )
            )

        _instrs_to_preload = []
        for i in instruments:
            if not fetcher.has_instrument(i):
                idx = fetcher.attach_instrument(i)
                self._instruments[idx] = (i, fetcher, subscription)  # type: ignore
                _instrs_to_preload.append(i)

        if self.is_running and _instrs_to_preload:
            self._slicer_ctrl += fetcher.load(
                pd.Timestamp(self._current_time, unit="ns"),  # type: ignore
                self._stop,  # type: ignore
                _instrs_to_preload,
            )

    def peek_historical_data(self, instrument: Instrument, subscription: str) -> list[Timestamped]:
        """
        Retrieves historical data for a specified instrument and subscription type up to the current simulation time.

        Parameters:
            - instrument (Instrument): instrument for which historical data is requested.
            - subscription (str): type of data subscription (e.g., OHLC, trades, quotes) for the instrument.

        Returns:
            - list[Timestamped]: A list of historical data elements for the specified instrument and subscription type
            that occurred before the current simulation time. If the simulation is not running, returns an empty list.

        Raises:
            SimulationError: If the instrument does not have the specified subscription in the simulation data provider.
        """
        if not self.has_subscription(instrument, subscription):
            raise SimulationError(
                f"Instrument: {instrument} has no subscription: {subscription} in this simulation data provider"
            )

        if not self.is_running:
            return []

        _subt_key, _, _ = self._parse_subscription_spec(subscription)
        _i_key = self._subtyped_fetchers[_subt_key].get_instrument_index(instrument)

        assert self._slicer_ctrl is not None and self._current_time is not None

        # fetch historical data for current time
        return self._slicer_ctrl.fetch_before_time(_i_key, self._current_time)

    def get_instruments_for_subscription(self, subscription: str) -> list[Instrument]:
        if subscription == DataType.ALL:
            return list((i for i, *_ in self._instruments.values()))

        _subt_key, _, _ = self._parse_subscription_spec(subscription)
        if (fetcher := self._subtyped_fetchers.get(_subt_key)) is not None:
            return [self._instruments[k][0] for k in fetcher.get_instruments_indices()]

        return []

    def get_subscriptions_for_instrument(self, instrument: Instrument | None) -> list[str]:
        r = []
        for i, f, s in self._instruments.values():
            if instrument is not None:
                if i == instrument:
                    r.append(s)
            else:
                r.append(s)
        return list(set(r))

    def has_subscription(self, instrument: Instrument, subscription_type: str) -> bool:
        for i, f, s in self._instruments.values():
            if i == instrument and s == subscription_type:
                return True
        return False

    def remove_instruments_from_subscription(self, subscription: str, instruments: list[Instrument] | Instrument):
        def _remove_from_fetcher(_subt_key: str, instruments: list[Instrument]):
            fetcher = self._subtyped_fetchers.get(_subt_key)
            if not fetcher:
                logger.warning(f"No configured data fetcher for '{_subt_key}' subscription !")
                return

            _keys_to_remove = []
            for i in instruments:
                # - try to remove from data fetcher
                if idx := fetcher.remove_instrument(i):
                    if idx in self._instruments:
                        self._instruments.pop(idx)
                        _keys_to_remove.append(idx)

            # print("REMOVING FROM:", _keys_to_remove)
            if self.is_running and _keys_to_remove:
                self._slicer_ctrl.remove(_keys_to_remove)  # type: ignore

        instruments = instruments if isinstance(instruments, list) else [instruments]

        # - if we want to remove instruments from all subscriptions
        if subscription == DataType.ALL:
            _f_keys = list(self._subtyped_fetchers.keys())
            for s in _f_keys:
                _remove_from_fetcher(s, instruments)
            return

        _subt_key, _, _ = self._parse_subscription_spec(subscription)
        _remove_from_fetcher(_subt_key, instruments)

    @property
    def is_running(self) -> bool:
        return self._current_time is not None

    def create_iterable(self, start: str | pd.Timestamp, stop: str | pd.Timestamp) -> Iterator:
        self._start = pd.Timestamp(start)
        self._stop = pd.Timestamp(stop)
        self._current_time = None
        self._slicer_ctrl = IteratedDataStreamsSlicer()
        return self

    def __iter__(self) -> Iterator:
        assert self._start is not None
        self._current_time = int(pd.Timestamp(self._start).timestamp() * 1e9)
        _ct_timestap = pd.Timestamp(self._current_time, unit="ns")

        for f in self._subtyped_fetchers.values():
            logger.debug(
                f"  [<c>IteratedDataStreamsSlicer</c>] :: Preloading initial data for {f._fetcher_id} {self._start} : {self._stop} ..."
            )
            self._slicer_ctrl += f.load(_ct_timestap, self._stop, None)  # type: ignore

        self._slicing_iterator = iter(self._slicer_ctrl)
        return self

    def __next__(self) -> tuple[Instrument, str, Timestamped, bool]:  # type: ignore
        try:
            while data := next(self._slicing_iterator):  # type: ignore
                k, t, v = data
                instr, fetcher, subt = self._instruments[k]
                data_type = fetcher._producing_data_type
                _is_historical = False
                if t < self._current_time:  # type: ignore
                    _is_historical = True
                else:
                    # only update the current time if the event is not historical
                    self._current_time = t

                return instr, data_type, v, _is_historical
        except StopIteration as e:  # noqa: F841
            raise StopIteration



from dataclasses import dataclass
from operator import neg

import numpy as np
from sortedcontainers import SortedDict

from qubx import logger
from qubx.core.basics import (
    OPTION_FILL_AT_SIGNAL_PRICE,
    OPTION_SIGNAL_PRICE,
    OPTION_SKIP_PRICE_CROSS_CONTROL,
    OPTION_AVOID_STOP_ORDER_PRICE_VALIDATION,
    Deal,
    Instrument,
    ITimeProvider,
    Order,
    OrderSide,
    OrderType,
    OrderStatus,
    TransactionCostsCalculator,
    dt_64,
)
from qubx.core.exceptions import (
    ExchangeError,
    InvalidOrder,
    SimulationError,
)
from qubx.core.series import OrderBook, Quote, Trade, TradeArray


@dataclass
class SimulatedExecutionReport:
    instrument: Instrument
    timestamp: dt_64
    order: Order
    exec: Deal | None


class OrdersManagementEngine:
    """
    Orders Management Engine (OME) is a simple implementation of a management of orders for simulation of a limit order book.

    2025-06-02: Added support for deferred execution reports (mainly for stop orders). This handles following cases:
        - It's possible to send stop loss order (STOP_MARKET) in on_execution_report() from custom PositionsTracker class
        - This order may be executed immediately that can lead to calling of on_execution_report() again (when we are still in on_execution_report())
        - To avoid this, it emulate stop orders execution (when condition is met) and add deferred execution report to the list
        - Deferred executions then would be sent on next process_market_data() call
    """

    instrument: Instrument
    time_service: ITimeProvider
    active_orders: dict[str, Order]
    stop_orders: dict[str, Order]
    asks: SortedDict  # [float, list[str]]
    bids: SortedDict  # [float, list[str]]
    bbo: Quote | None  # - current best bid/ask order book
    __prev_bbo: Quote | None  # - previous best bid/ask order book
    __order_id: int
    __trade_id: int
    _fill_stops_at_price: bool
    _tick_size: float
    _last_update_time: dt_64
    _last_data_update_time_ns: int
    _deferred_exec_reports: list[SimulatedExecutionReport]

    def __init__(
        self,
        instrument: Instrument,
        time_provider: ITimeProvider,
        tcc: TransactionCostsCalculator,
        fill_stop_order_at_price: bool = False,  # emulate stop orders execution at order's exact limit price
        debug: bool = True,
    ) -> None:
        self.instrument = instrument
        self.time_service = time_provider
        self.tcc = tcc
        self.asks = SortedDict()
        self.bids = SortedDict(neg)
        self.active_orders = dict()
        self.stop_orders = dict()
        self.bbo = None
        self.__order_id = 100000
        self.__trade_id = 100000
        self._fill_stops_at_price = fill_stop_order_at_price
        self._tick_size = instrument.tick_size
        self._last_update_time = np.datetime64(0, "ns")
        self._last_data_update_time_ns = 0
        self._deferred_exec_reports = []

        if not debug:
            self._dbg = lambda message, **kwargs: None

    def _generate_order_id(self) -> str:
        self.__order_id += 1
        return "SIM-ORDER-" + self.instrument.symbol + "-" + str(self.__order_id)

    def _generate_trade_id(self) -> str:
        self.__trade_id += 1
        return "SIM-EXEC-" + self.instrument.symbol + "-" + str(self.__trade_id)

    def get_quote(self) -> Quote:
        return self.bbo  # type: ignore

    def get_open_orders(self) -> list[Order]:
        return list(self.active_orders.values()) + list(self.stop_orders.values())

    def __remove_pending_status(self, exec: SimulatedExecutionReport) -> SimulatedExecutionReport:
        if exec.order.status == "PENDING":
            exec.order.status = "CLOSED"
        return exec

    def process_market_data(self, mdata: Quote | OrderBook | Trade | TradeArray) -> list[SimulatedExecutionReport]:
        """
        Processes the new market data (quote, trade or trades array) and simulates the execution of pending orders.
        """
        timestamp = self.time_service.time()
        _exec_report = []

        # - process deferred exec reports: spit out deferred exec reports in first place
        if self._deferred_exec_reports:
            _exec_report = [self.__remove_pending_status(i) for i in self._deferred_exec_reports]
            self._deferred_exec_reports.clear()

        # - pass through data if it's older than previous update
        if mdata.time < self._last_data_update_time_ns:
            return _exec_report
        self._last_data_update_time_ns = mdata.time
        _mkt_state = "UNKNOWN"

        # - new quote
        if isinstance(mdata, Quote):
            _b, _a = mdata.bid, mdata.ask
            _bs, _as = _b, _a
            _mkt_state = "Q: " + str(mdata)

            # - update BBO by new quote
            self.__prev_bbo = self.bbo
            self.bbo = mdata

        # - bunch of trades
        elif isinstance(mdata, TradeArray):
            # - to prevent triggering of orders on past trades in array
            _, max_buy_price, min_sell_price, _ = mdata.traded_range_from(self._last_update_time)
            _b = max_buy_price - self._tick_size
            _a = min_sell_price + self._tick_size
            _bs, _as = _a, _b
            _mkt_state = "TA: " + str(mdata)

        # - single trade
        elif isinstance(mdata, Trade):
            _b, _a = mdata.price - self._tick_size, mdata.price + self._tick_size
            _bs, _as = _b, _a
            _mkt_state = "T: " + str(mdata)

        # - order book
        elif isinstance(mdata, OrderBook):
            _b, _a = mdata.top_bid, mdata.top_ask
            _bs, _as = _b, _a
            _mkt_state = "OB: " + str(mdata)

        else:
            raise SimulationError(f"Invalid market data type: {type(mdata)} for update OME({self.instrument.symbol})")

        # - when new quote bid is higher than the lowest ask order execute all affected orders
        if self.asks and _b >= self.asks.keys()[0]:
            _asks_to_execute = list(self.asks.irange(0, _b))
            for level in _asks_to_execute:
                for order_id in self.asks[level]:
                    order = self.active_orders.pop(order_id)
                    _exec_report.append(self._execute_order(timestamp, order.price, order, False, _mkt_state))
                self.asks.pop(level)

        # - when new quote ask is lower than the highest bid order execute all affected orders
        if self.bids and _a <= self.bids.keys()[0]:
            _bids_to_execute = list(self.bids.irange(np.inf, _a))
            for level in _bids_to_execute:
                for order_id in self.bids[level]:
                    order = self.active_orders.pop(order_id)
                    _exec_report.append(self._execute_order(timestamp, order.price, order, False, _mkt_state))
                self.bids.pop(level)

        # - processing stop orders
        for soid in list(self.stop_orders.keys()):
            so = self.stop_orders[soid]
            _emulate_price_exec = self._fill_stops_at_price or so.options.get(OPTION_FILL_AT_SIGNAL_PRICE, False)

            if so.side == "BUY" and _as >= so.price:
                _exec_price = _as if not _emulate_price_exec else so.price
                self.stop_orders.pop(soid)
                _exec_report.append(self._execute_order(timestamp, _exec_price, so, True, _mkt_state))

            elif so.side == "SELL" and _bs <= so.price:
                _exec_price = _bs if not _emulate_price_exec else so.price
                self.stop_orders.pop(soid)
                _exec_report.append(self._execute_order(timestamp, _exec_price, so, True, _mkt_state))

        self._last_update_time = timestamp
        return _exec_report

    def place_order(
        self,
        order_side: OrderSide,
        order_type: OrderType,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        time_in_force: str = "gtc",
        **options,
    ) -> SimulatedExecutionReport:
        if self.bbo is None:
            raise ExchangeError(f"Simulator is not ready for order management - no quote for {self.instrument.symbol}")

        # - validate order parameters
        self._validate_order(order_side, order_type, amount, price, time_in_force, options)

        timestamp = self.time_service.time()
        order = Order(
            self._generate_order_id(),
            order_type,
            self.instrument,
            timestamp,
            amount,
            price if price is not None else 0,
            order_side,
            "NEW",
            time_in_force,
            client_id,
            options=options,
        )

        return self._process_order(timestamp, order)

    def _dbg(self, message, **kwargs) -> None:
        logger.debug(f"    [<y>OME</y>(<g>{self.instrument}</g>)] :: {message}", **kwargs)

    def _process_order(self, timestamp: dt_64, order: Order) -> SimulatedExecutionReport:
        if order.status in ["CLOSED", "CANCELED"]:
            raise InvalidOrder(f"Order {order.id} is already closed or canceled.")

        _buy_side = order.side == "BUY"
        _c_ask = self.bbo.ask  # type: ignore
        _c_bid = self.bbo.bid  # type: ignore

        # - check if order can be "executed" immediately
        _exec_price = None
        _need_update_book = False

        match order.type:
            case "MARKET":
                if _exec_price is None:
                    _exec_price = _c_ask if _buy_side else _c_bid

                # - special case only for simulation: exact fill at signal price for market orders
                _fill_at_signal_price = order.options.get(OPTION_FILL_AT_SIGNAL_PRICE, False)
                _signal_price = order.options.get(OPTION_SIGNAL_PRICE, None)

                # - some cases require to skip price cross control
                _skip_price_cross_control = order.options.get(OPTION_SKIP_PRICE_CROSS_CONTROL, False)

                # - it's passed only if signal price is valid: market crossed this desired price on last update
                if _fill_at_signal_price and _signal_price and self.__prev_bbo:
                    _desired_fill_price = _signal_price
                    _prev_mp = self.__prev_bbo.mid_price()
                    _c_mid_price = self.bbo.mid_price()  # type: ignore

                    if (
                        _skip_price_cross_control
                        or (_prev_mp < _desired_fill_price <= _c_mid_price)
                        or (_prev_mp > _desired_fill_price >= _c_mid_price)
                    ):
                        _exec_price = _desired_fill_price
                    else:
                        raise SimulationError(
                            f"Special execution price at {_desired_fill_price} for market order {order.id} cannot be filled because market didn't cross this price on last update !"
                        )

            case "LIMIT":
                _need_update_book = True
                if (_buy_side and order.price >= _c_ask) or (not _buy_side and order.price <= _c_bid):
                    _exec_price = _c_ask if _buy_side else _c_bid

            case "STOP_MARKET":
                # - it processes stop orders separately without adding to orderbook (as on real exchanges)
                order.status = "OPEN"
                _stp_order = order
                _emulate_price_exec = self._fill_stops_at_price or _stp_order.options.get(
                    OPTION_FILL_AT_SIGNAL_PRICE, False
                )

                if _stp_order.side == "BUY" and _c_ask >= _stp_order.price:
                    # _exec_price = _c_ask if not _emulate_price_exec else so.price
                    self._deferred_exec_reports.append(
                        self._execute_order(
                            timestamp,
                            _c_ask if not _emulate_price_exec else _stp_order.price,
                            order,
                            True,
                            "BBO: " + str(self.bbo),
                            "PENDING",
                        )
                    )

                elif _stp_order.side == "SELL" and _c_bid <= _stp_order.price:
                    # _exec_price = _c_bid if not _emulate_price_exec else so.price
                    self._deferred_exec_reports.append(
                        self._execute_order(
                            timestamp,
                            _c_bid if not _emulate_price_exec else _stp_order.price,
                            order,
                            True,
                            "BBO: " + str(self.bbo),
                            "PENDING",
                        )
                    )

                else:
                    self.stop_orders[order.id] = order

            case "STOP_LIMIT":
                # TODO: (OME) check trigger conditions in options etc
                raise NotImplementedError("'STOP_LIMIT' order is not supported in Qubx simulator yet !")

            case _:
                raise SimulationError(f"Invalid order type: {order.type} for {self.instrument.symbol}")

        # - if order must be "executed" immediately
        if _exec_price is not None:
            return self._execute_order(timestamp, _exec_price, order, True, "BBO: " + str(self.bbo))

        # - processing limit orders
        if _need_update_book:
            if _buy_side:
                self.bids.setdefault(order.price, list()).append(order.id)
            else:
                self.asks.setdefault(order.price, list()).append(order.id)

            order.status = "OPEN"
            self.active_orders[order.id] = order

        self._dbg(f"registered {order.id} {order.type} {order.side} {order.quantity} {order.price}")
        return SimulatedExecutionReport(self.instrument, timestamp, order, None)

    def _execute_order(
        self,
        timestamp: dt_64,
        exec_price: float,
        order: Order,
        taker: bool,
        market_state: str,
        status: OrderStatus = "CLOSED",
    ) -> SimulatedExecutionReport:
        order.status = status
        self._dbg(
            f"<red>{order.id}</red> {order.type} {order.side} {order.quantity} executed at {exec_price} ::: {market_state} [{status}]"
        )
        return SimulatedExecutionReport(
            self.instrument,
            timestamp,
            order,
            Deal(
                id=self._generate_trade_id(),
                order_id=order.id,
                time=timestamp,
                amount=order.quantity if order.side == "BUY" else -order.quantity,
                price=exec_price,
                aggressive=taker,
                fee_amount=self.tcc.get_execution_fees(
                    instrument=self.instrument, exec_price=exec_price, amount=order.quantity, crossed_market=taker
                ),
                fee_currency=self.instrument.quote,
            ),
        )

    def _validate_order(
        self, order_side: str, order_type: str, amount: float, price: float | None, time_in_force: str, options: dict
    ) -> None:
        if order_side.upper() not in ["BUY", "SELL"]:
            raise InvalidOrder("Invalid order side. Only BUY or SELL is allowed.")

        _ot = order_type.upper()
        if _ot not in ["LIMIT", "MARKET", "STOP_MARKET", "STOP_LIMIT"]:
            raise InvalidOrder("Invalid order type. Only LIMIT, MARKET, STOP_MARKET, STOP_LIMIT are supported.")

        if amount <= 0:
            raise InvalidOrder("Invalid order amount. Amount must be positive.")

        if (_ot == "LIMIT" or _ot.startswith("STOP")) and (price is None or price <= 0):
            raise InvalidOrder("Invalid order price. Price must be positively defined for LIMIT or STOP orders.")

        if time_in_force.upper() not in ["GTC", "IOC", "GTX"]:
            raise InvalidOrder("Invalid time in force. Only GTC, IOC, GTX are supported for now.")

        if _ot.startswith("STOP"):
            # - if the option is set, we don't check the current market price against the stop price
            if options.get(OPTION_AVOID_STOP_ORDER_PRICE_VALIDATION, False):
                return

            assert self.bbo
            assert price
            c_ask, c_bid = self.bbo.ask, self.bbo.bid
            if (order_side == "BUY" and c_ask >= price) or (order_side == "SELL" and c_bid <= price):
                raise ExchangeError(
                    f"Stop price would trigger immediately: STOP_MARKET {order_side} {amount} of {self.instrument.symbol} at {price} | market: {c_ask} / {c_bid}"
                )

    def cancel_order(self, order_id: str) -> SimulatedExecutionReport | None:
        # - check limit orders
        if order_id in self.active_orders:
            order = self.active_orders.pop(order_id)
            if order.side == "BUY":
                oids = self.bids[order.price]
                oids.remove(order_id)
                if not oids:
                    self.bids.pop(order.price)
            else:
                oids = self.asks[order.price]
                oids.remove(order_id)
                if not oids:
                    self.asks.pop(order.price)
        # - check stop orders
        elif order_id in self.stop_orders:
            order = self.stop_orders.pop(order_id)
        # - wrong order_id
        else:
            logger.error(f"Can't cancel order {order_id} for {self.instrument.symbol} because it's not found in OME !")
            return None

        order.status = "CANCELED"
        self._dbg(f"{order.id} {order.type} {order.side} {order.quantity} canceled")
        return SimulatedExecutionReport(self.instrument, self.time_service.time(), order, None)

    def __str__(self) -> str:
        _a, _b = True, True

        timestamp = self.time_service.time()
        _s = f"= = ({np.datetime64(timestamp, 'ns')}) = =\n"
        for k, v in reversed(self.asks.items()):
            _sizes = ",".join([f"{self.active_orders[o].quantity}" for o in v])
            _s += f"  {k} : [{_sizes}]\n"
            if k == self.bbo.ask:
                _a = False

        if _a:
            _s += f"  {self.bbo.ask} : \n"
        _s += "- - - - - - - - - - - - - - - - - - - -\n"

        _s1 = ""
        for k, v in self.bids.items():
            _sizes = ",".join([f"{self.active_orders[o].quantity}" for o in v])
            _s1 += f"  {k} : [{_sizes}]\n"
            if k == self.bbo.bid:
                _b = False
        _s1 += "= = = = = = = = = = = = = = = = = = = =\n"

        _s1 = f"  {self.bbo.bid} : \n" + _s1 if _b else _s1

        return _s + _s1



import ast
import fnmatch
import glob
import os
import shutil
import subprocess
import sys
import zipfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import toml
import yaml

from qubx import logger
from qubx.utils.misc import green


@dataclass
class PyClassInfo:
    name: str
    path: str
    docstring: str
    parameters: dict[str, Any]
    is_strategy: bool


def strategies_root() -> Path:
    """
    Get the root directory for strategies.
    """
    user_home = Path.home()
    strats_dir = user_home / "strategies"
    if not strats_dir.exists():
        strats_dir.mkdir()
    return strats_dir


def search_file(file_path: str | Path) -> Path | None:
    """
    Search for a file in the current directory and all parent directories.
    """
    file_path = Path(file_path)
    if file_path.exists():
        return file_path
    elif file_path.name == str(file_path):  # if only file name is provided
        curr_dir = Path.cwd()
        file_path = curr_dir / file_path
        while not file_path.exists():
            if curr_dir == curr_dir.parent:
                return None
            curr_dir = curr_dir.parent
            file_path = curr_dir / file_path.name
        return file_path
    return None


def zipdir(path: str | Path, zipname: str | Path) -> None:
    def _zipdir(path, ziph):
        for root, dirs, files in os.walk(path):
            for file_name in files:
                ziph.write(
                    os.path.join(root, file_name),
                    os.path.relpath(os.path.join(root, file_name), os.path.join(path, "..")),
                )

    logger.debug(zipname)
    with zipfile.ZipFile(f"{zipname}.zip", "w", zipfile.ZIP_DEFLATED) as zipf:
        _zipdir(path, zipf)


def flatten(decl_list: list):
    for d in decl_list:
        try:
            yield d.id
        except AttributeError:
            try:
                yield d.func.id
            except AttributeError:
                yield None


def is_decorated(decorator_list, decorator_name) -> bool:
    for x in flatten(decorator_list):
        if x == decorator_name:
            return True
    return False


def copy_to_dir(src: str, dest_dir: str):
    try:
        d = Path(dest_dir) / src
        os.makedirs(os.path.dirname(d), exist_ok=True)
        shutil.copy2(src, d)
        logger.debug(f"  Copying <g>{src}</g> -> <c>{d}</c>")
    except:  # noqa: E722
        logger.error(f" - Error copying {src} -> {d}")


def copy_file_to_dir(src: str, dest_dir: str, new_name: str | None = None):
    try:
        if new_name:
            d = Path(dest_dir) / new_name
        else:
            d = Path(dest_dir) / os.path.split(src)[-1]
        os.makedirs(os.path.dirname(d), exist_ok=True)
        shutil.copy2(src, d)
        logger.debug(f"  Copying <g>{src}</g> -> <c>{d}</c>")
    except:  # noqa: E722
        logger.error(f" - Error copying {src} -> {d}")


def ask_y_n(question: str) -> bool:
    user_input = input(f" -> {green(question)} (yes/no, default=yes): ")
    return user_input.lower() in ["yes", "y", ""]


def get_current_python_version() -> str:
    """
    Retrieves the current Python version (major.minor) as a string.
    """
    try:
        output = subprocess.check_output(["python", "--version"], text=True)
        version = output.strip().split()[1]
        major_minor = ".".join(version.split(".")[:2])  # Extract major.minor
        return major_minor
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Error retrieving Python version: {e}")


def find_file_by_mask(mask: str | Path, root_dir: str | Path) -> list[Path]:
    """
    Finds files matching the given mask starting from the root directory.

    Args:
        mask (str): The file mask to search for (e.g., "qubx/utils/runner.py").
        root_dir (str): The root directory to start the search.

    Returns:
        list: A list of full paths to files matching the mask.
    """
    matches = []
    for root, dirnames, filenames in os.walk(root_dir):
        for filename in fnmatch.filter(filenames, os.path.basename(mask)):
            if mask in os.path.join(root, filename):  # TODO: WTF ???
                matches.append(Path(os.path.join(root, filename)))
    return matches


def scan_py_classes_in_directory(
    root_path: str | Path, skip_directories: tuple[str, ...] = ("dist",)
) -> list[PyClassInfo]:
    """
    Scan Python classes in the given directory. Additionaly it will extract parameters from the class and mark it as a strategy.

    Args:
        root_path: The root directory to scan for classes
        skip_directories: Tuple of directory names to skip during scanning. Defaults to ("dist",)

    Returns:
        A list of PyClassInfo objects.
    """
    classes = []
    root_path = os.path.expanduser(root_path)

    for file_path in glob.glob(os.path.join(root_path, "**/*.py"), recursive=True):
        # Skip files in directories that should be skipped
        if any(skip_dir in file_path.split(os.sep) for skip_dir in skip_directories):
            continue

        name = os.path.splitext(os.path.basename(file_path))[0]

        # Ignore __ files
        if name.startswith("__"):
            continue

        with open(file_path, "r") as f:
            src = f.read()

        try:
            class_node = ast.parse(src)
        except:  # noqa: E722
            continue

        nodes = [node for node in ast.walk(class_node) if isinstance(node, ast.ClassDef)]
        for n in nodes:
            if len(n.bases) > 0:
                for b in n.bases:
                    if hasattr(b, "id") and getattr(b, "id", None) in ["IStrategy"]:
                        # - when class inherits IStrategy, it is a strategy
                        classes.append(
                            PyClassInfo(
                                name=n.name,
                                path=file_path,
                                docstring=ast.get_docstring(n) or "",
                                parameters=iter_class_parameters(n),
                                is_strategy=True,
                            )
                        )
                        break
                else:
                    # - otherwise we just add it as a class but still extract parameters etc
                    classes.append(
                        PyClassInfo(
                            name=n.name,
                            path=file_path,
                            docstring=ast.get_docstring(n) or "",
                            parameters=iter_class_parameters(n),
                            is_strategy=False,
                        )
                    )
                    continue
                break
            else:
                classes.append(
                    PyClassInfo(
                        name=n.name,
                        path=file_path,
                        docstring=ast.get_docstring(n) or "",
                        parameters=iter_class_parameters(n),
                        is_strategy=False,
                    )
                )
    return classes


def iter_class_parameters(class_node: ast.AST) -> dict[str, Any]:
    """
    Iterate over the parameters of a class.
    """
    decls = {}
    for n in ast.iter_child_nodes(class_node):
        if isinstance(n, ast.AnnAssign):
            if not n.target.id.startswith("_"):
                if hasattr(n, "value"):
                    try:
                        decls[n.target.id] = None
                        if n.value:
                            if isinstance(n.value.value, ast.Name):
                                decls[n.target.id] = "???"
                            else:
                                decls[n.target.id] = n.value.value

                    except Exception as _:
                        decls[n.target.id] = None
    return dict(reversed(list(decls.items())))


def find_runner_file(root_path: str | Path) -> Path:
    """
    Find qubx runner file in the given root directory.
    """
    if runner_files := find_file_by_mask(mask="qubx/utils/runner.py", root_dir=Path(root_path)):
        assert len(runner_files) == 1, f"Multiple runner files found in {root_path}"
        return runner_files[0]
    raise FileNotFoundError(f"Runner file not found in {root_path}")


def find_strategy_data(strategy_name: str, start_dir: str | Path) -> tuple[Path | None, Path | None, Path | None]:
    """
    Find root directory of the strategy by its configuration files / directories
    """
    start_dir = Path(start_dir).expanduser()
    assert start_dir.exists(), f"Start directory {start_dir} does not exist"

    for root, dirs, files in os.walk(start_dir):
        for file_name in files:
            if file_name.endswith(".yaml") or file_name.endswith(".yml"):
                with open(os.path.join(root, file_name), "r") as f:
                    try:
                        yaml_data = yaml.safe_load(f)
                        if yaml_data["config"]["strategy"].split(".")[-1].lower() == strategy_name.lower():
                            if runner_files := find_file_by_mask(mask="qubx/utils/runner.py", root_dir=Path(root)):
                                assert len(runner_files) == 1, f"Multiple runner files found in {root}"
                                return Path(root), runner_files[0], Path(os.path.join(root, file_name))
                            else:
                                logger.error(f"Runner qubx file not found in {root}. Skipping...")
                    except Exception as e:
                        logger.error(f"Error processing {file_name}: {str(e)}")
    return None, None, None


def find_git_root(file_path: str) -> str:
    """
    Find the root directory of a Git repository from a file path.

    :param file_path: Path to a file within the repository.
    :return: The root directory of the Git repository.
    :raises ValueError: If the file is not inside a Git repository.
    """
    try:
        dir_path = os.path.abspath(os.path.dirname(file_path))

        git_root = subprocess.check_output(
            ["git", "-C", dir_path, "rev-parse", "--show-toplevel"], stderr=subprocess.STDOUT, text=True
        ).strip()

        return git_root
    except subprocess.CalledProcessError as e:
        raise ValueError(f"The file '{file_path}' is not inside a Git repository.") from e


def find_pyproject_root(file_path: str) -> str:
    """
    Find the root directory of a pyproject.toml file from a file path.
    """
    try:
        dir_path = os.path.abspath(os.path.dirname(file_path))
        while not os.path.exists(os.path.join(dir_path, "pyproject.toml")):
            dir_path = os.path.dirname(dir_path)
        return dir_path
    except Exception as e:
        raise ValueError(f"The file '{file_path}' is not inside a pyproject.toml file.") from e


def locate_poetry(paths: list[str] | None = None, silent: bool = False) -> Path | None:
    """
    Locate the Poetry executable.

    :param paths: Optional list of paths to search for the Poetry executable.
    :return: The path to the Poetry executable.
    :raises FileNotFoundError: If the Poetry executable is not found.
    """
    poetry_executable = shutil.which("poetry") or shutil.which("poetry.exe")
    if poetry_executable is None:
        poetry_paths = [
            Path.home() / ".poetry" / "bin" / "poetry",
            Path.home() / ".local" / "bin" / "poetry",
            Path("/usr/local/bin/poetry"),
            Path("/usr/bin/poetry"),
        ]
        poetry_executable = next(
            (path for path in poetry_paths if path.is_file() or Path(f"{path}.exe").is_file()), None
        )

    if not poetry_executable:
        if not silent:
            raise FileNotFoundError("Poetry executable not found. Ensure Poetry is installed and in the PATH.")
        else:
            logger.error("Poetry executable not found. Ensure Poetry is installed and in the PATH.")
            return None
    return poetry_executable


def _is_package_platform_compatible(package: dict[str, Any]) -> bool:
    files = package.get("files", [])
    current_platform = sys.platform
    if "win" in current_platform:
        current_platforms = ["win32", "win_amd64", "win_arm64", "any"]
    elif "linux" in current_platform:
        current_platforms = ["linux", "any"]
    else:
        current_platforms = ["any"]
    for file_data in files:
        if file_data["file"].endswith(".whl"):
            if any(platform in file_data["file"] for platform in current_platforms):
                return True
    return len(files) == 0


def generate_dependency_file(
    lock_file: str | Path = "poetry.lock",
    output_file: str | Path = "pyproject.toml",
    project_name: str = "generated-project",
    save_as_requirements: bool = False,
):
    try:
        with open(lock_file, "r") as lock:
            lock_data = toml.load(lock)

        if save_as_requirements:
            requirements = []
            for package in lock_data.get("package", []):
                dep_name = package["name"]
                version = package["version"]
                if not _is_package_platform_compatible(package):
                    continue
                requirements.append(f"{dep_name}=={version}")

            with open(output_file, "w") as requirements_file:
                requirements_file.write("\n".join(requirements))

            print(f"requirements.txt has been generated and saved to {output_file}.")
        else:
            pyproject = {
                "tool": {
                    "poetry": {
                        "name": project_name,
                        "version": "0.1.0",
                        "description": "Generated from poetry.lock",
                        "dependencies": {},
                        "dev-dependencies": {},
                    }
                }
            }

            for package in lock_data.get("package", []):
                dep_name = package["name"]
                version = package["version"]
                category = package.get("category", "main")
                if not _is_package_platform_compatible(package):
                    continue

                if category == "main":
                    pyproject["tool"]["poetry"]["dependencies"][dep_name] = version
                elif category == "dev":
                    pyproject["tool"]["poetry"]["dev-dependencies"][dep_name] = version

            with open(output_file, "w") as pyproject_file:
                toml.dump(pyproject, pyproject_file)

            logger.debug(f"pyproject.toml has been generated and saved to {output_file}.")
    except Exception as e:
        print(f"An error occurred during making dependencies file: {e}")



from typing import Any

from qubx import logger
from qubx.core.basics import Instrument, MarketType, Order, OrderRequest, OrderSide
from qubx.core.interfaces import IAccountProcessor, IBroker, ITimeProvider, ITradingManager


class ClientIdStore:
    """Manages generation of unique client order IDs."""

    def __init__(self):
        """Initialize a client ID store."""
        self._order_id: int | None = None

    def generate_id(self, time_provider: ITimeProvider, symbol: str) -> str:
        """Generate a unique client order ID.

        Args:
            time_provider: Time provider to get current timestamp
            symbol: Trading symbol for the order

        Returns:
            A unique client order ID
        """
        # Initialize order ID from timestamp if not yet set
        if self._order_id is None:
            self._order_id = self._initialize_id_from_timestamp(time_provider)

        # Increment order ID to ensure uniqueness across calls
        self._order_id += 1

        # Create and return the unique ID
        return self._create_id(symbol, self._order_id)

    def _initialize_id_from_timestamp(self, time_provider: ITimeProvider) -> int:
        """Initialize the order ID from the current timestamp.

        Args:
            time_provider: Time provider to get current timestamp

        Returns:
            Initial order ID value
        """
        return time_provider.time().astype("int64") // 100_000_000

    def _create_id(self, symbol: str, order_id: int) -> str:
        """Create the ID from symbol and order ID.

        Args:
            symbol: Trading symbol
            order_id: Current order ID counter

        Returns:
            Client ID string
        """
        return "_".join(["qubx", symbol, str(order_id)])


class TradingManager(ITradingManager):
    _time_provider: ITimeProvider
    _brokers: list[IBroker]
    _account: IAccountProcessor
    _strategy_name: str

    _client_id_store: ClientIdStore
    _exchange_to_broker: dict[str, IBroker]

    def __init__(
        self, time_provider: ITimeProvider, brokers: list[IBroker], account: IAccountProcessor, strategy_name: str
    ) -> None:
        self._time_provider = time_provider
        self._brokers = brokers
        self._account = account
        self._strategy_name = strategy_name
        self._client_id_store = ClientIdStore()
        self._exchange_to_broker = {broker.exchange(): broker for broker in brokers}

    def trade(
        self,
        instrument: Instrument,
        amount: float,
        price: float | None = None,
        time_in_force="gtc",
        client_id: str | None = None,
        **options,
    ) -> Order:
        size_adj = self._adjust_size(instrument, amount)
        side = self._get_side(amount)
        type = self._get_order_type(instrument, price, options)
        price = self._adjust_price(instrument, price, amount)
        client_id = client_id or self._generate_order_client_id(instrument.symbol)

        logger.debug(
            f"[<g>{instrument.symbol}</g>] :: Sending (blocking) {type} {side} {size_adj} {' @ ' + str(price) if price else ''} -> (client_id: <r>{client_id})</r> ..."
        )

        order = self._exchange_to_broker[instrument.exchange].send_order(
            instrument=instrument,
            order_side=side,
            order_type=type,
            amount=size_adj,
            price=price,
            time_in_force=time_in_force,
            client_id=client_id,
            **options,
        )

        if order is not None:
            self._account.add_active_orders({order.id: order})

        return order

    def trade_async(
        self,
        instrument: Instrument,
        amount: float,
        price: float | None = None,
        time_in_force="gtc",
        client_id: str | None = None,
        **options,
    ) -> None:
        size_adj = self._adjust_size(instrument, amount)
        side = self._get_side(amount)
        type = self._get_order_type(instrument, price, options)
        price = self._adjust_price(instrument, price, amount)
        client_id = client_id or self._generate_order_client_id(instrument.symbol)

        logger.debug(
            f"[<g>{instrument.symbol}</g>] :: Sending (async) {type} {side} {size_adj} {' @ ' + str(price) if price else ''} -> (client_id: <r>{client_id})</r> ..."
        )

        self._exchange_to_broker[instrument.exchange].send_order_async(
            instrument=instrument,
            order_side=side,
            order_type=type,
            amount=size_adj,
            price=price,
            time_in_force=time_in_force,
            client_id=client_id,
            **options,
        )

    def submit_orders(self, order_requests: list[OrderRequest]) -> list[Order]:
        raise NotImplementedError("Not implemented yet")

    def set_target_position(
        self, instrument: Instrument, target: float, price: float | None = None, time_in_force="gtc", **options
    ) -> Order:
        raise NotImplementedError("Not implemented yet")

    def set_target_leverage(
        self, instrument: Instrument, leverage: float, price: float | None = None, **options
    ) -> Order:
        raise NotImplementedError("Not implemented yet")

    def close_position(self, instrument: Instrument) -> None:
        raise NotImplementedError("Not implemented yet")

    def close_positions(self, market_type: MarketType | None = None) -> None:
        raise NotImplementedError("Not implemented yet")

    def cancel_order(self, order_id: str, exchange: str | None = None) -> None:
        if not order_id:
            return
        if exchange is None:
            exchange = self._brokers[0].exchange()
        self._exchange_to_broker[exchange].cancel_order(order_id)
        self._account.remove_order(order_id, exchange)

    def cancel_orders(self, instrument: Instrument) -> None:
        for o in self._account.get_orders(instrument).values():
            self.cancel_order(o.id, instrument.exchange)

    def _generate_order_client_id(self, symbol: str) -> str:
        return self._client_id_store.generate_id(self._time_provider, symbol)

    def exchanges(self) -> list[str]:
        return list(self._exchange_to_broker.keys())

    def _adjust_size(self, instrument: Instrument, amount: float) -> float:
        size_adj = instrument.round_size_down(abs(amount))
        if size_adj < instrument.min_size:
            raise ValueError(
                f"[{instrument.symbol}] Attempt to trade size {abs(amount)} less than minimal allowed {instrument.min_size} !"
            )
        return size_adj

    def _adjust_price(self, instrument: Instrument, price: float | None, amount: float) -> float | None:
        if price is None:
            return price
        return instrument.round_price_down(price) if amount > 0 else instrument.round_price_up(price)

    def _get_side(self, amount: float) -> OrderSide:
        return "BUY" if amount > 0 else "SELL"

    def _get_order_type(self, instrument: Instrument, price: float | None, options: dict[str, Any]) -> str:
        if price is None:
            return "market"
        if (stp_type := options.get("stop_type")) is not None:
            return f"stop_{stp_type}"
        return "limit"



from typing import Any

import pandas as pd

from qubx.core.basics import AssetType, Instrument, MarketType


def ccxt_build_qubx_exchange_name(ccxt_exchange: str, market_type: str | None = None, is_linear: bool = True) -> str:
    """
    Build a Qubx exchange name from a ccxt exchange name and a market dictionary.
    Parameters:
        ccxt_exchange (str): The ccxt exchange name.
        market (dict): The market dictionary.
    Returns:
        str: The Qubx exchange name.
    """
    if ccxt_exchange == "BINANCE.PM":
        if market_type in ["spot", "margin"]:
            return "BINANCE"
        elif market_type == "swap" and is_linear:
            return "BINANCE.UM"
        elif market_type == "swap" and not is_linear:
            return "BINANCE.CM"
        else:
            return "BINANCE.UM"
    else:
        # for not just return the input exchange and extend later if needed
        return ccxt_exchange


def ccxt_symbol_to_instrument(ccxt_exchange_name: str, market: dict[str, Any]) -> Instrument:
    exchange = ccxt_build_qubx_exchange_name(ccxt_exchange_name, market["type"], market.get("linear", True))
    inner_info = market["info"]
    maint_margin = 0.0
    required_margin = 0.0
    liquidation_fee = 0.0
    if "marginLevels" in inner_info:
        margins = inner_info["marginLevels"][0]
        maint_margin = float(margins["maintenanceMargin"])
        required_margin = float(margins["initialMargin"])
    if "maintMarginPercent" in inner_info:
        maint_margin = float(inner_info["maintMarginPercent"]) / 100
    if "requiredMarginPercent" in inner_info:
        required_margin = float(inner_info["requiredMarginPercent"]) / 100
    if "liquidationFee" in inner_info:
        liquidation_fee = float(inner_info["liquidationFee"])

    # symbol = market["id"]
    # let's use unified symbol format across all exchanges / types: BASEQUOTE
    symbol = market["base"] + market["quote"]

    # add some exchange specific formatting of symbol name
    tick_size = float(market["precision"]["price"] or 0.0)
    lot_size = float(market["precision"]["amount"] or 0.0)
    min_size = float(market["limits"]["amount"]["min"] or 0.0)
    min_notional = float(market["limits"]["cost"]["min"] or 0.0)

    if exchange in ["BITFINEX", "BITFINEX.F"]:
        if symbol.startswith("t"):
            symbol = symbol[1:]
        symbol = symbol.replace(":", "-")
        tick_size = 10**-tick_size
        lot_size = 10**-lot_size
        min_size = 10**-min_size
        min_notional = 10**-min_notional

    mkt_type = MarketType[market["type"].upper()]

    # - extract expiry date if present
    expiry_date = pd.Timestamp(market["expiryDatetime"]) if "expiryDatetime" in market else None
    if not expiry_date and "expiry" in market:
        expiry_date = pd.Timestamp(int(market["expiry"]), unit="ms") if "expiry" in market else None

    # - add expiry date to futures symbol if present
    if mkt_type == MarketType.FUTURE and expiry_date:
        symbol += f".{expiry_date.strftime('%Y%m%d')}"

    return Instrument(
        symbol=symbol,
        asset_type=AssetType.CRYPTO,
        market_type=mkt_type,
        exchange=exchange,
        base=market["base"],
        quote=market["quote"],
        settle=market["settle"],
        exchange_symbol=market["id"],
        tick_size=tick_size,
        lot_size=lot_size,
        min_size=min_size,
        min_notional=min_notional,
        initial_margin=required_margin,
        maint_margin=maint_margin,
        liquidation_fee=liquidation_fee,
        contract_size=float(market.get("contractSize", 1.0) or 1.0),
        onboard_date=pd.Timestamp(int(inner_info["onboardDate"]), unit="ms") if "onboardDate" in inner_info else None,
        delivery_date=expiry_date,
        inverse=market.get("inverse", False),
    )


def ccxt_fetch_instruments(
    exchange_to_ccxt_name: dict[str, str],
    keep_types: list[MarketType] | None = None,
    instruments: dict[str, Instrument] | None = None,
) -> dict[str, Instrument]:
    """
    Fetch instruments from CCXT.

    Parameters:
        exchange_to_ccxt_name (dict[str, str]): The exchange to CCXT name mapping.
        keep_types (list[MarketType] | None): The market types to keep.
        instruments (dict[str, Instrument] | None): The instruments to update.
    Returns:
        dict[str, Instrument]: The updated instruments.
    """
    import ccxt as cx

    # - make a copy of the instruments dict
    instruments = {} if instruments is None else dict(instruments)

    # - replace defaults with data from CCXT
    for exch, ccxt_name in exchange_to_ccxt_name.items():
        exch = exch.upper()
        ccxt_name = ccxt_name.lower()
        ex = getattr(cx, ccxt_name)()
        mkts = ex.load_markets()
        for v in mkts.values():
            if v["index"]:
                continue
            instr = ccxt_symbol_to_instrument(exch, v)
            if not keep_types or instr.market_type in keep_types:
                instruments[str(instr)] = instr

    # - return updated instruments
    return instruments



from typing import TYPE_CHECKING, Any, Callable, Dict, Type, TypeVar

from qubx.utils.misc import class_import

if TYPE_CHECKING:
    from qubx.data.readers import DataReader

    T = TypeVar("T", bound="DataReader")
else:
    # Use Any as a placeholder during runtime
    T = TypeVar("T", bound=Any)


class ReaderRegistry:
    """
    A registry for data readers that allows registering and retrieving reader classes by name.

    This registry is used to map reader names (like 'mqdb', 'csv', etc.) to their respective
    reader classes, allowing for dynamic lookup and instantiation of readers.
    """

    _readers: Dict[str, Type[Any]] = {}

    @classmethod
    def register(cls, name: str) -> Callable[[Type[T]], Type[T]]:
        """
        Decorator to register a reader class with the registry.

        Args:
            name: The name to register the reader under

        Returns:
            A decorator function that registers the class
        """

        def decorator(reader_cls: Type[T]) -> Type[T]:
            cls._readers[name] = reader_cls
            return reader_cls

        return decorator

    @classmethod
    def get(cls, reader_name: str, **kwargs) -> Any:
        """
        Get a reader instance by name.

        This method handles both URI-style names (e.g., 'mqdb::nebula') and
        fully qualified class names (e.g., 'sty.data.readers.MyCustomDataReader').

        Args:
            reader_name: The name of the reader to retrieve, can be a URI or class name
            **kwargs: Additional arguments to pass to the reader constructor

        Returns:
            An instance of the reader

        Raises:
            ValueError: If the reader cannot be found or instantiated
        """
        _is_uri = "::" in reader_name

        if _is_uri:
            # Handle URI-style names like 'mqdb::nebula' or 'csv::/data/rawdata/'
            db_conn, db_name = reader_name.split("::")

            # Try to get the reader from the registry
            reader_cls = cls._readers.get(db_conn)
            if reader_cls is not None:
                # Different readers have different parameter names for the connection string
                # For CSV readers, it's 'path', for database readers, it's 'host'
                return reader_cls(db_name, **kwargs)  # type: ignore  # noqa
        else:
            # Check if it's a simple name registered in the registry
            reader_cls = cls._readers.get(reader_name)
            if reader_cls is not None:
                return reader_cls(**kwargs)

            # Try to import the class directly
            try:
                reader_cls = class_import(reader_name)
                return reader_cls(**kwargs)
            except (ImportError, AttributeError, ValueError):
                pass

        # If we get here, we couldn't find or instantiate the reader
        raise ValueError(f"Unknown reader type: {reader_name}")

    @classmethod
    def is_registered(cls, name: str) -> bool:
        """
        Check if a reader is registered with the given name.

        Args:
            name: The name to check

        Returns:
            True if the reader is registered, False otherwise
        """
        return name in cls._readers

    @classmethod
    def get_all_readers(cls) -> Dict[str, Type[Any]]:
        """
        Get all registered readers.

        Returns:
            A dictionary mapping reader names to reader classes
        """
        return cls._readers.copy()


def reader(name: str) -> Callable[[Type[T]], Type[T]]:
    """
    Decorator for registering a reader class with the registry.

    This is a convenience function that calls ReaderRegistry.register.

    Args:
        name: The name to register the reader under

    Returns:
        A decorator function that registers the class
    """
    return ReaderRegistry.register(name)



"""
Composite Metric Emitter.

This module provides a composite implementation of IMetricEmitter that delegates to multiple emitters.
"""

from typing import Dict, List, Optional

from qubx import logger
from qubx.core.basics import dt_64
from qubx.core.interfaces import IMetricEmitter, IStrategyContext
from qubx.emitters.base import BaseMetricEmitter


class CompositeMetricEmitter(BaseMetricEmitter):
    """
    A composite metric emitter that delegates to multiple emitters.

    This emitter can be used to send metrics to multiple monitoring systems at once.
    """

    def __init__(
        self,
        emitters: List[IMetricEmitter],
        stats_to_emit: Optional[List[str]] = None,
        stats_interval: str = "1m",
        tags: Optional[Dict[str, str]] = None,
    ):
        """
        Initialize the Composite Metric Emitter.

        Args:
            emitters: List of emitters to delegate to
            stats_to_emit: Optional list of specific stats to emit
            stats_interval: Interval for emitting strategy stats (default: "1m")
            tags: Dictionary of default tags/labels to include with all metrics
        """
        super().__init__(stats_to_emit, stats_interval, tags)
        self._emitters = emitters

    def _emit_impl(self, name: str, value: float, tags: Dict[str, str], timestamp: dt_64 | None = None) -> None:
        """
        Implementation of emit for the composite emitter.

        Args:
            name: Name of the metric
            value: Value of the metric
            tags: Dictionary of tags/labels for the metric (already merged with default tags)
            timestamp: Optional timestamp for the metric
        """
        for emitter in self._emitters:
            try:
                emitter.emit(name, value, tags, timestamp)
            except Exception as e:
                logger.error(f"Error emitting metric to {emitter.__class__.__name__}: {e}")

    def emit_strategy_stats(self, context: IStrategyContext) -> None:
        """
        Emit standard strategy statistics to all configured emitters.

        Args:
            context: The strategy context to get statistics from
        """
        # Store context for later use in notify method
        self._context = context

        for emitter in self._emitters:
            try:
                emitter.emit_strategy_stats(context)
            except Exception as e:
                logger.error(f"Error emitting strategy stats to {emitter.__class__.__name__}: {e}")

    def notify(self, context: IStrategyContext) -> None:
        for emitter in self._emitters:
            try:
                emitter.notify(context)
            except Exception as e:
                logger.error(f"Error notifying {emitter.__class__.__name__}: {e}")



"""
Utility functions for the restorers module.
"""

import re
from pathlib import Path
from typing import Optional


def find_latest_run_folder(base_dir: str | Path) -> Optional[Path]:
    """
    Find the latest run folder in the given directory.

    The function looks for folders matching the pattern 'run_YYYYMMDDHHMMSS'
    and returns the path to the most recent one.

    Args:
        base_dir: The base directory to search in.

    Returns:
        The path to the latest run folder, or None if no run folders are found.
    """
    base_path = Path(base_dir)
    if not base_path.exists() or not base_path.is_dir():
        return None

    # Find all directories matching the run pattern
    run_pattern = re.compile(r"run_\d{14}")
    run_folders = []

    for item in base_path.iterdir():
        if item.is_dir() and run_pattern.match(item.name):
            run_folders.append(item)

    if not run_folders:
        return None

    # Sort by folder name (which contains the timestamp)
    # The latest run will have the highest timestamp
    latest_run = sorted(run_folders, key=lambda x: x.name, reverse=True)[0]

    return latest_run



from collections import defaultdict
from typing import Iterable

import numpy as np
import pandas as pd

from ccxt.pro import Exchange
from qubx import logger
from qubx.core.basics import DataType, Instrument
from qubx.data.readers import DataReader, DataTransformer
from qubx.data.registry import reader
from qubx.utils.misc import AsyncThreadLoop
from qubx.utils.time import handle_start_stop

from .factory import get_ccxt_exchange
from .utils import ccxt_find_instrument, instrument_to_ccxt_symbol


@reader("ccxt")
class CcxtDataReader(DataReader):
    SUPPORTED_DATA_TYPES = {"ohlc"}

    _exchanges: dict[str, Exchange]
    _loop: AsyncThreadLoop
    _max_bars: int = 10_000
    _max_history: pd.Timedelta

    def __init__(
        self,
        exchanges: list[str],
        max_bars: int = 10_000,
        max_history: str = "30d",
    ):
        _exchanges = [e.upper() for e in exchanges]
        _loop = None
        self._exchanges = {}
        for exchange in _exchanges:
            self._exchanges[exchange] = get_ccxt_exchange(exchange, loop=_loop)
            if _loop is None and hasattr(self._exchanges[exchange], "asyncio_loop"):
                _loop = getattr(self._exchanges[exchange], "asyncio_loop")

        assert _loop is not None, "Loop is not set"
        self._loop = AsyncThreadLoop(_loop)

        self._max_bars = max_bars
        self._max_history = pd.Timedelta(max_history)
        self._exchange_to_symbol_to_instrument = defaultdict(dict)

        _tasks = []
        for exchange in self._exchanges:
            _tasks.append(self._loop.submit(self._exchanges[exchange].load_markets()))
        for task in _tasks:
            task.result()

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        timeframe: str = "1m",
        data_type: str = DataType.OHLC,
        **kwargs,
    ) -> Iterable | list:
        if data_type not in self.SUPPORTED_DATA_TYPES:
            return []

        instrument = self._get_instrument(data_id)
        if instrument is None:
            return []

        _timeframe = pd.Timedelta(timeframe or "1m")
        _start, _stop = self._get_start_stop(start, stop, _timeframe)

        if _start > _stop:
            return []

        ccxt_symbol = instrument_to_ccxt_symbol(instrument)
        exchange = self._exchanges[instrument.exchange]
        data = self._fetch_ohlcv(ccxt_symbol, timeframe, _start, _stop, exchange)
        column_names = self._get_column_names(data_type)

        if not chunksize:
            transform.start_transform(data_id, column_names, start=start, stop=stop)
            transform.process_data(data)
            return transform.collect()

        def _iter_chunks():
            for i in range(0, len(data), chunksize):
                chunk = data[i : i + chunksize]
                transform.start_transform(data_id, column_names, start=start, stop=stop)
                transform.process_data(chunk)
                yield transform.collect()

        return _iter_chunks()

    def get_names(self, **kwargs) -> list[str]:
        return list(self._exchanges.keys())

    def get_symbols(self, exchange: str, dtype: str) -> list[str]:
        if dtype not in self.SUPPORTED_DATA_TYPES:
            return []

        exchange = exchange.upper()
        if exchange not in self._exchanges:
            return []

        _exchange = self._exchanges[exchange]
        markets = _exchange.markets
        if markets is None:
            return []

        symbols = list(markets.keys())

        instruments = [
            ccxt_find_instrument(symbol, _exchange, self._exchange_to_symbol_to_instrument[exchange])
            for symbol in symbols
        ]
        return [f"{exchange}:{instrument.symbol}" for instrument in instruments]

    def get_time_ranges(self, symbol: str, dtype: str) -> tuple[np.datetime64 | None, np.datetime64 | None]:
        if dtype != "ohlc":
            return None, None

        end_time = pd.Timestamp.now()
        start_time = end_time - self._max_history
        return start_time.to_datetime64(), end_time.to_datetime64()

    def close(self):
        tasks = []
        for exchange in self._exchanges:
            if hasattr(self._exchanges[exchange], "close") and callable(self._exchanges[exchange].close):
                tasks.append(self._loop.submit(self._exchanges[exchange].close()))

        for task in tasks:
            try:
                task.result(timeout=5)  # Wait up to 5 seconds for the connection to close
            except Exception as e:
                logger.error(f"Error closing exchange connection: {e}")

    def _get_instrument(self, data_id: str) -> Instrument | None:
        parts = data_id.split(":")
        if len(parts) != 2:
            return None
        exchange, symbol = parts
        exchange = exchange.upper()
        if exchange not in self._exchanges:
            return None
        return ccxt_find_instrument(symbol, self._exchanges[exchange], self._exchange_to_symbol_to_instrument[exchange])

    def _get_start_stop(
        self, start: str | None, stop: str | None, timeframe: pd.Timedelta
    ) -> tuple[pd.Timestamp, pd.Timestamp]:
        if not stop:
            stop = pd.Timestamp.now().isoformat()
        _start, _stop = handle_start_stop(start, stop, convert=lambda x: pd.Timestamp(x))
        assert isinstance(_stop, pd.Timestamp)
        if not _start:
            _start = _stop - timeframe * self._max_bars
        assert isinstance(_start, pd.Timestamp)

        if _start < (_max_time := pd.Timestamp.now() - self._max_history):
            _start = _max_time

        return _start, _stop

    def _fetch_ohlcv(
        self, symbol: str, timeframe: str, start: pd.Timestamp, stop: pd.Timestamp, exchange: Exchange
    ) -> list:
        since = int(start.timestamp() * 1000)
        until = int(stop.timestamp() * 1000)

        future = self._loop.submit(self._async_fetch_ohlcv(symbol, timeframe, since, until, exchange))

        try:
            ohlcv = future.result()
            return [
                [
                    pd.Timestamp(candle[0], unit="ms").to_pydatetime(),
                    float(candle[1]),  # open
                    float(candle[2]),  # high
                    float(candle[3]),  # low
                    float(candle[4]),  # close
                    float(candle[5]),  # volume
                ]
                for candle in ohlcv
            ]
        except Exception as e:
            logger.error(f"Error fetching OHLCV data: {e}")
            return []

    async def _async_fetch_ohlcv(self, symbol: str, timeframe: str, since: int, until: int, exchange: Exchange) -> list:
        all_candles = []

        # CCXT has a limit on how many candles can be fetched at once
        # We need to fetch in batches
        limit = 1000
        current_since = since

        while True:
            try:
                candles = await exchange.fetch_ohlcv(symbol, timeframe, since=current_since, limit=limit)

                if not candles:
                    break

                all_candles.extend(candles)

                # Update the since parameter for the next batch
                last_timestamp = candles[-1][0]

                # If we've reached the until timestamp, stop fetching
                if until and last_timestamp >= until:
                    # Filter out candles after the until timestamp
                    all_candles = [c for c in all_candles if c[0] <= until]
                    break

                # If we got fewer candles than the limit, we've reached the end
                if len(candles) < limit:
                    break

                # Set the since parameter to the timestamp of the last candle + 1
                current_since = last_timestamp + 1

            except Exception as e:
                logger.error(f"Error fetching OHLCV data: {e}")
                break

        return all_candles

    def _get_column_names(self, data_type: str) -> list[str]:
        match data_type:
            case DataType.OHLC:
                return ["timestamp", "open", "high", "low", "close", "volume"]
            case _:
                return []



import re
import zipfile
from collections import defaultdict
from pathlib import Path

import numpy as np
import pandas as pd
import yaml

from qubx.core.metrics import TradingSessionResult
from qubx.utils.misc import blue, cyan, green, magenta, red, yellow


class BacktestsResultsManager:
    """
    Manager class for handling backtesting results.

    This class provides functionality to load, list and manage backtesting results stored in zip files.
    Each result contains trading session information and metrics that can be loaded and analyzed.

    Parameters
    ----------
    path : str
        Path to directory containing backtesting result zip files

    Methods
    -------
    - reload()
        Reloads all backtesting results from the specified path
    - list(regex="", with_metrics=False)
        Lists all backtesting results, optionally filtered by regex and including metrics
    - load(name)
        Loads a specific backtesting result by name
    - load_config(name)
        Loads the configuration YAML file for a specific backtest result
    - delete(name)
        Deletes one or more backtest results
    """

    def __init__(self, path: str):
        self.path = path
        self.reload()

    def reload(self) -> "BacktestsResultsManager":
        self.results = {}
        self.variations = {}

        _vars = defaultdict(list)
        names = defaultdict(lambda: 0)
        for p in Path(self.path).glob("**/*.zip"):
            with zipfile.ZipFile(p, "r") as zip_ref:
                try:
                    info = yaml.safe_load(zip_ref.read("info.yml"))
                    info["path"] = str(p)
                    n = info.get("name", "")
                    var_set_name = info.get("variation_name", "")

                    # - put variations aside
                    if var_set_name:
                        _vars[var_set_name].append(info)
                        continue

                    _new_name = n if names[n] == 0 else f"{n}.{names[n]}"
                    names[n] += 1
                    info["name"] = _new_name
                    self.results[_new_name] = info
                except Exception:
                    pass

        # - reindex
        _idx = 1
        for n in sorted(self.results.keys()):
            self.results[n]["idx"] = _idx
            _idx += 1

        # - reindex variations at the end
        for n in sorted(_vars.keys()):
            self.variations[_idx] = {
                "name": n,
                "idx": _idx,
                "variations": _vars[n],
                "created": pd.Timestamp(_vars[n][0].get("creation_time", "")).round("1s"),
                "author": _vars[n][0].get("author", ""),
                "description": _vars[n][0].get("description", ""),
            }
            _idx += 1

        return self

    def __getitem__(
        self, name: str | int | list[int] | list[str] | slice
    ) -> TradingSessionResult | list[TradingSessionResult]:
        return self.load(name)

    def load(
        self, name_or_idx: str | int | list[int] | list[str] | slice
    ) -> TradingSessionResult | list[TradingSessionResult]:
        match name_or_idx:
            case list():
                return [self.load(i) for i in name_or_idx]  # type: ignore
            case str():
                return [self.load(i) for i in self._find_indices(name_or_idx)]  # type: ignore
            case slice():
                return [
                    self.load(i)
                    for i in range(name_or_idx.start, name_or_idx.stop, name_or_idx.step if name_or_idx.step else 1)
                ]  # type: ignore
            case int():
                if name_or_idx > len(self.results) and name_or_idx in self.variations:
                    return [
                        TradingSessionResult.from_file(v.get("path", ""))
                        for v in self.variations[name_or_idx].get("variations", [])
                    ]

                # - load by index
                for info in self.results.values():
                    if info.get("idx", -1) == name_or_idx:
                        return TradingSessionResult.from_file(info["path"])

        raise ValueError(f"No result found for '{name_or_idx}' !")

    def load_config(self, name: str | int) -> str:
        """Load the configuration YAML file for a specific backtest result.

        Args:
            name (str | int): The name or index of the backtest result. If str, matches against the backtest name.
                            If int, matches against the backtest index.

        Returns:
            str: The contents of the configuration YAML file as a string.

        Raises:
            ValueError: If no backtest result is found matching the provided name/index.
        """
        p = None
        for info in self.results.values():
            match name:
                case int():
                    if info.get("idx", -1) == name:
                        n = info.get("name", "")
                        p = info.get("path", {})
                        break
                case str():
                    if info.get("name", "") == name:
                        n = info.get("name", "")
                        p = info.get("path", {})
                        break
        if p is None:
            raise ValueError(f"No result found for {name}")

        # - name may have .1, .2, etc. so we need to remove it
        n = n.split(".")[0] if "." in n else n
        with zipfile.ZipFile(p, "r") as zip_ref:
            return zip_ref.read(f"{n}.yaml").decode("utf-8")

    def delete(self, name: str | int | list[int] | list[str] | slice):
        """Delete one or more backtest results.

        Args:
            name: Identifier(s) for the backtest result(s) to delete. Can be:
                - str: Name of backtest or regex pattern to match multiple backtests
                - int: Index of specific backtest
                - list[int]: List of backtest indices
                - list[str]: List of backtest names
                - slice: Range of backtest indices to delete

        Prints:
            Message confirming which backtest(s) were deleted, or error if none found.
            Deleted backtest names are shown in red text.

        Note:
            - For string names, supports regex pattern matching against backtest names and strategy class names
            - Deletes the underlying results files and reloads the results index
            - Operation is irreversible
        """

        def _del_idx(idx):
            for info in self.results.values():
                if info.get("idx", -1) == idx:
                    Path(info["path"]).unlink()
                    return info.get("name", idx)
            return None

        match name:
            case str():
                nms = [_del_idx(i) for i in self._find_indices(name)]
                self.reload()
                print(f" -> Deleted {red(', '.join(nms))} ...")
                return

            case list():
                nms = [_del_idx(i) for i in name]
                self.reload()
                print(f" -> Deleted {red(', '.join(nms))} ...")
                return

            case slice():
                nms = [_del_idx(i) for i in range(name.start, name.stop, name.step if name.step else 1)]
                self.reload()
                print(f" -> Deleted {red(', '.join(nms))} ...")
                return

        for info in self.results.values():
            match name:
                case int():
                    if info.get("idx", -1) == name:
                        Path(info["path"]).unlink()
                        print(f" -> Deleted {red(info.get('name', name))} ...")
                        self.reload()
                        return
                case str():
                    if info.get("name", "") == name:
                        Path(info["path"]).unlink()
                        print(f" -> Deleted {red(info.get('name', name))} ...")
                        self.reload()
                        return
        print(f" -> No results found for {red(name)} !")

    def _find_indices(self, regex: str):
        for n in sorted(self.results.keys()):
            info = self.results[n]
            s_cls = info.get("strategy_class", "").split(".")[-1]

            try:
                if not re.match(regex, n, re.IGNORECASE):
                    # if not re.match(regex, s_cls, re.IGNORECASE):
                    continue
            except Exception:
                if regex.lower() != n.lower() and regex.lower() != s_cls.lower():
                    continue

            yield info.get("idx", -1)

    def list_variations(self, regex: str = "", detailed=True, sort_by: str | None = "sharpe", ascending=False):
        """
        List only variations of a backtest result.

        Args:
            - regex (str, optional): Regular expression pattern to filter results by strategy name or class. Defaults to "".
            - sort_by (str, optional): The criterion to sort the results by. Defaults to "sharpe".
            - ascending (bool, optional): Whether to sort the results in ascending order. Defaults to False.
            - detailed (bool, optional): Whether to show each variation run. Defaults to True.
        """
        return self.list(
            regex=regex,
            sort_by=sort_by,
            ascending=ascending,
            show_variations=True,
            show_simulations=False,
            show_each_variation_run=detailed,
        )

    def list(
        self,
        regex: str = "",
        with_metrics=True,
        params=False,
        as_table=False,
        pretty_print=False,
        sort_by: str | None = "sharpe",
        ascending=False,
        show_simulations=True,
        show_variations=True,
        show_each_variation_run=True,
    ):
        """List backtesting results with optional filtering and formatting.

        Args:
            - regex (str, optional): Regular expression pattern to filter results by strategy name or class. Defaults to "".
            - with_metrics (bool, optional): Whether to include performance metrics in output. Defaults to True.
            - params (bool, optional): Whether to display strategy parameters. Defaults to False.
            - as_table (bool, optional): Return results as a pandas DataFrame instead of printing. Defaults to False.
            - sort_by (str, optional): The criterion to sort the results by. Defaults to "sharpe".
            - ascending (bool, optional): Whether to sort the results in ascending order. Defaults to False.
            - show_simulations (bool, optional): Whether to show simulation results. Defaults to True.
            - show_variations (bool, optional): Whether to show variation results. Defaults to True.
            - show_each_variation_run (bool, optional): Whether to show each variation run. Defaults to True.

        Returns:
            - Optional[pd.DataFrame]: If as_table=True, returns a DataFrame containing the results sorted by creation time.
            - Otherwise prints formatted results to console.
        """
        _t_rep = []
        if show_simulations:
            for n in sorted(self.results.keys()):
                info = self.results[n]
                s_cls = info.get("strategy_class", "").split(".")[-1]

                if regex:
                    if not re.match(regex, n, re.IGNORECASE):
                        # if not re.match(regex, s_cls, re.IGNORECASE):
                        continue

                name = info.get("name", "")
                smbs = ", ".join(info.get("symbols", list()))
                start = pd.Timestamp(info.get("start", "")).round("1s")
                stop = pd.Timestamp(info.get("stop", "")).round("1s")
                dscr = info.get("description", "")
                created = pd.Timestamp(info.get("creation_time", "")).round("1s")
                metrics = info.get("performance", {})
                author = info.get("author", "")
                _s = f"{yellow(str(info.get('idx')))} - {red(name)} ::: {magenta(created)} by {cyan(author)}"

                _one_line_dscr = ""
                if dscr:
                    dscr = dscr.split("\n")
                    for _d in dscr:
                        _s += f"\n\t{magenta('# ' + _d)}"
                        _one_line_dscr += "\u25cf " + _d + "\n"

                _s += f"\n\tstrategy: {green(s_cls)}"
                _s += f"\n\tinterval: {blue(start)} - {blue(stop)}"
                _s += f"\n\tcapital: {blue(info.get('capital', ''))} {info.get('base_currency', '')} ({info.get('commissions', '')})"
                _s += f"\n\tinstruments: {blue(smbs)}"
                if params:
                    formats = ["{" + f":<{i}" + "}" for i in [50]]
                    _p = pd.DataFrame.from_dict(info.get("parameters", {}), orient="index")
                    for i in _p.to_string(
                        max_colwidth=30,
                        header=False,
                        formatters=[(lambda x: cyan(fmt.format(str(x)))) for fmt in formats],
                        justify="left",
                    ).split("\n"):
                        _s += f"\n\t  |  {yellow(i)}"

                if not as_table:
                    print(_s)

                if with_metrics:
                    _m_repr = (
                        pd.DataFrame.from_dict(metrics, orient="index")
                        .T[["gain", "cagr", "sharpe", "qr", "max_dd_pct", "mdd_usd", "fees", "execs"]]
                        .astype(float)
                    )
                    _m_repr = _m_repr.round(3).to_string(index=False)
                    _h, _v = _m_repr.split("\n")
                    if not as_table:
                        print("\t " + red(_h))
                        print("\t " + cyan(_v))

                if not as_table:
                    print()
                else:
                    metrics = {
                        m: round(v, 3)
                        for m, v in metrics.items()
                        if m in ["gain", "cagr", "sharpe", "qr", "max_dd_pct", "mdd_usd", "fees", "execs"]
                    }
                    _t_rep.append(
                        {"Index": info.get("idx", ""), "Strategy": name}
                        | metrics
                        | {
                            "start": start,
                            "stop": stop,
                            "Created": created,
                            "Author": author,
                            "Description": _one_line_dscr,
                        },
                    )

        # - variations (only if not as_table for the time being)
        if not as_table and show_variations:
            for _i, vi in self.variations.items():
                n = vi.get("name", "")
                if regex:
                    if not re.match(regex, n, re.IGNORECASE):
                        continue

                _s = f"{yellow(str(_i))} - {red(str(n))} set of {len(vi.get('variations'))} variations ::: {magenta(vi.get('created'))} by {cyan(vi.get('author'))}"

                dscr = vi.get("description", "").split("\n")
                for _d in dscr:
                    _s += f"\n\t{magenta('# ' + _d)}"

                _mtrx = {}
                for v in vi.get("variations", []):
                    _nm = v.get("name", "")
                    _nm = _nm.split("_")[-1].strip("()")
                    _mtrx[_nm] = v.get("performance", {})

                _m_repr = pd.DataFrame.from_dict(_mtrx, orient="index")[
                    ["gain", "cagr", "sharpe", "qr", "max_dd_pct", "mdd_usd", "fees", "execs"]
                ].astype(float)
                _m_repr = _m_repr.round(3)
                _m_repr = _m_repr.sort_values(by=sort_by, ascending=ascending) if sort_by else _m_repr
                _m_repr = _m_repr.to_string(index=True)

                print(_s)
                if show_each_variation_run:
                    for _i, _l in enumerate(_m_repr.split("\n")):
                        if _i == 0:
                            print("\t " + red(_l))
                        else:
                            print("\t " + blue(_l))

        if as_table:
            _df = pd.DataFrame.from_records(_t_rep, index="Index")
            _df = _df.sort_values(by=sort_by, ascending=ascending) if sort_by else _df
            if pretty_print:
                from IPython.display import HTML

                return HTML(
                    _df.to_html()
                    .replace("\\n", "<br><hr style='border-color: #005000; '/>")
                    .replace("<td>", '<td align="left" valign="top">')
                )
            return _df

    def variation_plot(self, variation_idx: int, criterion: str = "sharpe", ascending: bool = False, n=3, h=600):
        """
        Plot a variation of a backtest result.

        Args:
            - variation_idx (int): The index of the variation to plot.
            - criterion (str): The criterion to plot (e.g. "sharpe", "mdd_usd", "max_dd_pct", etc.).
            - ascending (bool): Whether to sort the results in ascending order.
            - n (int): The number of decimal places to display.
            - h (int): The height of the plot.

        Returns:
            plotly.graph_objects.Figure: The plot of the variation.
        """
        import plotly.express as px
        from itertools import cycle
        from qubx.utils.misc import string_shortener

        _vars = self.variations.get(variation_idx)
        if not _vars:
            raise ValueError(f"No variations found for index {variation_idx} !")

        variations = _vars.get("variations", [])
        name = _vars.get("name", "") or ""

        _r, _p = {}, {}
        for i, v in enumerate(variations):
            _p[i] = v["parameters"]
        _pp = pd.DataFrame.from_records(_p).T
        # - changed parameters
        _cp = []
        for c in _pp.columns:
            if len(_pp[c].astype(str).unique()) > 1:
                _cp.append(c)

        # - if nothing was actually changed in parameters, raise an error
        if not _cp:
            raise ValueError(f"No variable parameters found for simulation {name} !")

        _ms = max([len(string_shortener(x)) for x in _cp]) + 3
        _h = "".join([string_shortener(x).center(_ms) for x in _cp])

        _sel = lambda ds, _cp: "".join(
            [
                f"<span style='color:{c}'> {str(ds[k]).center(_ms)}</span>"
                for k, c in zip(_cp, cycle(px.colors.qualitative.Plotly))
                if k in k in ds
            ]
        )
        for i, v in enumerate(variations):
            _r[i] = {"name": v["name"], **v["performance"], "parameters": _sel(v["parameters"], _cp)}

        t1 = pd.DataFrame.from_records(_r).T
        if criterion not in t1.columns:
            raise ValueError(f"Criterion {criterion} not found in results: possible values are {t1.columns}")
        t2 = t1.sort_values(criterion, ascending=ascending)

        data = pd.Series([np.nan, *t2[criterion].to_list()], index=[_h, *t2["parameters"].to_list()])

        figure = (
            px.bar(data, orientation="h")
            .update_layout(
                title=dict(
                    text=f"{name} | <span style='color:orange'>{criterion.capitalize()}</span>",
                ),
                xaxis=dict(tickfont=dict(family="monospace", size=10, color="#ff4000")),
                yaxis=dict(
                    tickfont=dict(family="monospace", size=10, color="#40a000"),
                    dtick=1,
                ),
            )
            .update_layout(
                height=h,
                hovermode="x unified",
                showlegend=False,
                hoverdistance=1,
                yaxis={"hoverformat": f".{n}f"},
                dragmode="zoom",
                newshape=dict(line_color="red", line_width=1.0),
                modebar_add=["drawline", "drawopenpath", "drawrect", "eraseshape"],
                hoverlabel=dict(align="auto", bgcolor="rgba(10, 10, 10, 0.5)"),
            )
            .update_xaxes(
                showspikes=True,
                spikemode="across",
                spikesnap="cursor",
                spikecolor="#306020",
                spikethickness=1,
                spikedash="dot",
                title=criterion,
            )
            .update_yaxes(
                spikesnap="cursor",
                spikecolor="#306020",
                tickformat=f".{n}f",
                spikethickness=1,
                title="Parameters",
                autorange="reversed",
            )
        )
        return figure



"""
Indicator Emitter Module.

This module provides the IndicatorEmitter class that can wrap around any indicator
and automatically emit their values when there are updates.
"""

import numpy as np

from qubx import logger
from qubx.core.basics import Instrument
from qubx.core.interfaces import IMetricEmitter
from qubx.core.series import Indicator
from qubx.core.utils import recognize_time


class IndicatorEmitter(Indicator):
    """
    An indicator that wraps around another indicator and automatically emits its values.

    This class solves the problem of wanting to emit indicator values independently
    of when certain calculation methods are called. For example, you might want to
    emit ATR values on every update, not just when calculate_risks() is called.

    Example:
        # Create ATR indicator
        volatility = atr(ohlc_series, period=14, smoother="sma")

        # Wrap it with an emitter to automatically emit values
        volatility_emitter = IndicatorEmitter.wrap_with_emitter(
            volatility,
            metric_emitter=my_emitter,
            metric_name="atr_volatility",
            instrument=instrument,
            tags={"timeframe": "1m", "period": "14"}
        )
    """

    def __init__(
        self,
        name: str,
        wrapped_indicator: Indicator,
        metric_emitter: IMetricEmitter,
        metric_name: str | None = None,
        instrument: Instrument | None = None,
        tags: dict[str, str] | None = None,
        emit_on_new_item_only: bool = True,
    ):
        """
        Initialize the IndicatorEmitter.

        Args:
            name: Name of this emitter indicator
            wrapped_indicator: The indicator to wrap and emit values from
            metric_emitter: The metric emitter to use for sending values
            metric_name: Name to use for the emitted metric
            instrument: Optional instrument to include in tags
            tags: Optional additional tags to include with emitted metrics
            emit_on_new_item_only: If True, only emit when new_item_started=True;
                                  if False, emit on every update
        """
        # Store our configuration BEFORE calling super().__init__()
        # because super().__init__() triggers initial data recalculation
        self._wrapped_indicator = wrapped_indicator
        self._metric_emitter = metric_emitter
        self._metric_name = metric_name if metric_name is not None else wrapped_indicator.name
        self._instrument = instrument
        self._tags = tags or {}
        self._emit_on_new_item_only = emit_on_new_item_only
        self._has_emitted = False

        # Initialize as an indicator subscribed to the wrapped indicator's updates
        # Since Indicator extends TimeSeries, we can pass the wrapped_indicator directly
        super().__init__(name, wrapped_indicator)

        logger.debug(
            f"[IndicatorEmitter] Created emitter '{name}' wrapping '{wrapped_indicator.name}' -> metric '{self._metric_name}'"
        )

    def calculate(self, time: int, value: float, new_item_started: bool) -> float:
        """
        Calculate method that handles the emission logic.

        Args:
            time: Timestamp of the update
            value: The input value (from the wrapped indicator - should be numeric)
            new_item_started: Whether this is a new item or an update to existing

        Returns:
            The current value of the wrapped indicator
        """
        current_value = float(value)
        if not np.isfinite(current_value):
            current_value = np.nan

        # Decide whether to emit based on our configuration
        should_emit = False

        if not self._emit_on_new_item_only:
            # Emit on every update
            should_emit = not np.isnan(current_value)
        else:
            # Only emit when new item starts AND we have a previous value to emit
            if new_item_started and len(self._wrapped_indicator) >= 2:
                should_emit = True
                # Use the previous (completed) value, not the current one
                current_value = self._wrapped_indicator[1] if not self.is_initial_recalculate else value

        # Emit if we should and the value is valid
        if should_emit and not np.isnan(current_value):
            try:
                # Prepare tags for emission
                emission_tags = self._tags.copy()

                # Emit the metric with the proper timestamp
                self._metric_emitter.emit(
                    name=self._metric_name,
                    value=float(current_value),
                    tags=emission_tags,
                    timestamp=recognize_time(time),
                    instrument=self._instrument,
                )

                if not self._has_emitted:
                    logger.debug(
                        f"[IndicatorEmitter] '{self.name}' started emitting '{self._metric_name}' "
                        f"values from '{self._wrapped_indicator.name}'"
                    )
                    self._has_emitted = True

            except Exception as e:
                logger.error(
                    f"[IndicatorEmitter] Failed to emit metric '{self._metric_name}' "
                    f"from '{self._wrapped_indicator.name}': {e}"
                )

        # Return the current value
        return current_value

    @classmethod
    def wrap_with_emitter(
        cls,
        indicator: Indicator,
        metric_emitter: IMetricEmitter,
        metric_name: str | None = None,
        instrument: Instrument | None = None,
        tags: dict[str, str] | None = None,
        emit_on_new_item_only: bool = True,
    ) -> "Indicator":
        """
        Convenience method to wrap an existing indicator with an emitter.

        Args:
            indicator: The indicator to wrap
            metric_emitter: The metric emitter to use
            metric_name: Name for the emitted metric (defaults to indicator name)
            instrument: Optional instrument for tagging
            tags: Optional additional tags
            emit_on_new_item_only: Whether to emit only on new items or on every update

        Returns:
            IndicatorEmitter: The wrapped indicator with emission capability
        """
        return cls.wrap(
            indicator,
            metric_emitter=metric_emitter,
            metric_name=metric_name,
            instrument=instrument,
            tags=tags,
            emit_on_new_item_only=emit_on_new_item_only,
        )

    def __getitem__(self, idx):
        """Delegate indexing to the wrapped indicator."""
        return self._wrapped_indicator[idx]

    def __len__(self):
        """Delegate length to the wrapped indicator."""
        return len(self._wrapped_indicator)


def indicator_emitter(
    wrapped_indicator: Indicator,
    metric_emitter: IMetricEmitter,
    metric_name: str | None = None,
    instrument: Instrument | None = None,
    tags: dict[str, str] | None = None,
    emit_on_new_item_only: bool = True,
) -> IndicatorEmitter:
    """
    Helper function to create an IndicatorEmitter following the standard pattern.

    Args:
        wrapped_indicator: The indicator to wrap
        metric_emitter: The metric emitter to use
        metric_name: Name for the emitted metric (defaults to indicator name)
        instrument: Optional instrument for tagging
        tags: Optional additional tags
        emit_on_new_item_only: Whether to emit only on new items or on every update

    Returns:
        IndicatorEmitter: The wrapped indicator with emission capability
    """
    emitter = IndicatorEmitter.wrap_with_emitter(
        indicator=wrapped_indicator,
        metric_emitter=metric_emitter,
        metric_name=metric_name,
        instrument=instrument,
        tags=tags,
        emit_on_new_item_only=emit_on_new_item_only,
    )
    assert isinstance(emitter, IndicatorEmitter)
    return emitter



import gzip
import os
import traceback
from collections import defaultdict
from datetime import datetime
from os.path import exists, join
from pathlib import Path
from typing import Any

import msgspec
import numpy as np
import pandas as pd
from numba import njit, types
from numba.typed import Dict
from tqdm.auto import tqdm

from qubx import QubxLogConfig, logger
from qubx.core.basics import Instrument
from qubx.core.lookups import lookup
from qubx.pandaz.utils import scols, srows
from qubx.utils.numbers_utils import count_decimal_places


@njit
def prec_floor(a: float, precision: int) -> float:
    return np.sign(a) * np.true_divide(np.floor(round(abs(a) * 10**precision, precision)), 10**precision)


@njit
def prec_ceil(a: float, precision: int):
    return np.sign(a) * np.true_divide(np.ceil(round(abs(a) * 10**precision, precision)), 10**precision)


@njit
def get_tick(price: float, is_bid: bool, tick_size: float):
    if is_bid:
        return int(np.floor(round(price / tick_size, 1)))
    else:
        return int(np.ceil(round(price / tick_size, 1)))


@njit
def tick_to_price(tick: int, tick_size: float, decimals: int):
    return round(tick * tick_size, decimals)


@njit
def get_tick_price(price: float, is_bid: bool, tick_size: float, decimals: int):
    return tick_to_price(get_tick(price, is_bid, tick_size), tick_size, decimals)


@njit
def _interpolate_levels(
    levels: list[tuple[float, float]],
    is_bid: bool,
    tick_count: int,
    tick_size: float,
    decimals: int,
    size_decimals: int,
    sizes_in_quoted: bool,
):
    # TODO: asks are not interpolated correctly
    prices = []
    for price, size in levels:
        prices.append(price)

    if is_bid:
        max_tick = get_tick(max(prices), is_bid, tick_size)
        min_tick = max_tick - tick_count + 1
        start_tick = max_tick
    else:
        min_tick = get_tick(min(prices), is_bid, tick_size)
        max_tick = min_tick + tick_count - 1
        start_tick = min_tick

    # Initialize a dictionary to hold the aggregated sizes
    interp_levels = Dict.empty(key_type=types.float64, value_type=types.float64)

    # Iterate through each bid and aggregate the sizes based on the tick size
    for price, size in levels:
        tick = get_tick(price, is_bid, tick_size)
        if tick >= min_tick and tick <= max_tick:
            _size = (price * size) if sizes_in_quoted else size
            if tick in interp_levels:
                interp_levels[tick] += _size
            else:
                interp_levels[tick] = _size

    # Create the final list including zero sizes where necessary
    result = []
    for tick in range(min_tick, max_tick + 1):
        size = round(interp_levels[tick], size_decimals) if tick in interp_levels else 0.0
        idx = tick - start_tick
        result.append((-idx if is_bid else idx, size))

    return result, tick_to_price(max_tick if is_bid else min_tick, tick_size, decimals)


@njit
def __build_orderbook_snapshots(
    dates: np.ndarray,
    prices: np.ndarray,
    sizes: np.ndarray,
    is_bids: np.ndarray,
    levels: int,
    tick_size_fraction: float,
    price_decimals: int,
    size_decimals: int,
    sizes_in_quoted: bool,
    init_bid_ticks: np.ndarray,
    init_bid_sizes: np.ndarray,
    init_ask_ticks: np.ndarray,
    init_ask_sizes: np.ndarray,
    init_top_bid: float,
    init_top_ask: float,
    init_tick_size: float,
) -> list[tuple[np.datetime64, list[tuple[float, float]], list[tuple[float, float]], float, float, float]]:
    """
    Build order book snapshots from given market data.

    Parameters:
        dates (np.ndarray): Array of datetime64 timestamps.
        prices (np.ndarray): Array of price points.
        sizes (np.ndarray): Array of sizes corresponding to the prices.
        is_bids (np.ndarray): Array indicating if the price is a bid (True) or ask (False).
        levels (int): Number of levels to interpolate for bids and asks.
        tick_size_fraction (float): Fraction to determine the tick size dynamically based on mid-price.
        price_decimals (int): Number of decimal places for price rounding.
        size_decimals (int): Number of decimal places for size rounding.
        sizes_in_quoted (bool): Flag indicating if sizes are in quoted currency.
        init_bid_ticks (np.ndarray): Initial bid ticks.
        init_bid_sizes (np.ndarray): Initial bid sizes.
        init_ask_ticks (np.ndarray): Initial ask ticks.
        init_ask_sizes (np.ndarray): Initial ask sizes.
        init_top_bid (float): Initial top bid price.
        init_top_ask (float): Initial top ask price.
        init_tick_size (float): Initial tick size.

    Returns:
    list[tuple[np.datetime64, list[tuple[float, float]], list[tuple[float, float]], float, float, float]]:
        A list of tuples where each tuple contains:
        - Timestamp of the snapshot.
        - List of interpolated bid levels (price, size).
        - List of interpolated ask levels (price, size).
        - Top bid price.
        - Top ask price.
        - Tick size.
    """
    price_to_size = Dict.empty(key_type=types.float64, value_type=types.float64)
    price_to_bid_ask = Dict.empty(key_type=types.float64, value_type=types.boolean)

    for i in range(init_bid_ticks.shape[0]):
        bp = init_top_bid - init_tick_size * init_bid_ticks[i]
        price_to_size[bp] = init_bid_sizes[i]
        price_to_bid_ask[bp] = True

    for i in range(init_ask_ticks.shape[0]):
        ap = init_top_ask + init_tick_size * init_ask_ticks[i]
        price_to_size[ap] = init_ask_sizes[i]
        price_to_bid_ask[ap] = False

    snapshots = []
    prev_timestamp = dates[0]
    for i in range(dates.shape[0]):
        date = dates[i]
        if date > prev_timestamp:
            # emit snapshot
            bids, asks = [], []
            top_a, top_b = np.inf, 0
            for price, size in price_to_size.items():
                if price_to_bid_ask[price]:
                    bids.append((price, size))
                    top_b = max(top_b, price)
                else:
                    asks.append((price, size))
                    top_a = min(top_a, price)

            if len(bids) > 0 and len(asks) > 0:
                # - find tick_size dynamically based on mid_price
                tick_size = prec_ceil(0.5 * (top_b + top_a) * tick_size_fraction, price_decimals)
                interp_bids, top_bid_price = _interpolate_levels(
                    bids,
                    True,
                    levels,
                    tick_size,
                    price_decimals,
                    size_decimals,
                    sizes_in_quoted,
                )
                interp_asks, top_ask_price = _interpolate_levels(
                    asks,
                    False,
                    levels,
                    tick_size,
                    price_decimals,
                    size_decimals,
                    sizes_in_quoted,
                )
                if len(interp_bids) >= levels and len(interp_asks) >= levels:
                    if top_bid_price <= top_ask_price:
                        snapshots.append(
                            (
                                prev_timestamp,
                                interp_bids[-levels:],
                                interp_asks[:levels],
                                # - also store top bid, ask prices and tick_size
                                top_b,
                                top_a,
                                tick_size,
                            )
                        )
                    else:
                        # something went wrong, bids can't be above asks
                        # clean up the local state and hope for the best
                        price_to_size.clear()
                        price_to_bid_ask.clear()

        price = prices[i]
        size = sizes[i]
        is_bid = is_bids[i]
        if size == 0:
            if price in price_to_size:
                del price_to_size[price]
            if price in price_to_bid_ask:
                del price_to_bid_ask[price]
        else:
            price_to_size[price] = size
            price_to_bid_ask[price] = is_bid

        prev_timestamp = date

    return snapshots


def build_orderbook_snapshots(
    updates: list[tuple[np.datetime64, float, float, bool]],
    levels: int,
    tick_size_pct: float,
    min_tick_size: float,
    min_size_step: float,
    sizes_in_quoted: bool = False,
    initial_snapshot: (
        tuple[
            np.datetime64,  # timestamp   [0]
            list[tuple[float, float]],  # bids levels [1]
            list[tuple[float, float]],  # asks levels [2]
            float,
            float,
            float,  # top bid, top ask prices, tick_size [3, 4, 5]
        ]
        | None
    ) = None,
):
    dates, prices, sizes, is_bids = zip(*updates)
    dates = np.array(dates, dtype=np.datetime64)
    prices = np.array(prices)
    sizes = np.array(sizes)
    is_bids = np.array(is_bids)

    price_decimals = max(count_decimal_places(min_tick_size), 1)
    size_decimals = max(count_decimal_places(min_size_step), 1)

    if initial_snapshot is not None and dates[0] > initial_snapshot[0]:
        init_bid_ticks, init_bid_sizes = zip(*initial_snapshot[1])
        init_ask_ticks, init_ask_sizes = zip(*initial_snapshot[2])
        init_bid_ticks = np.array(init_bid_ticks, dtype=np.float64)
        init_bid_sizes = np.array(init_bid_sizes, dtype=np.float64)
        init_ask_ticks = np.array(init_ask_ticks, dtype=np.float64)
        init_ask_sizes = np.array(init_ask_sizes, dtype=np.float64)
        init_top_bid = initial_snapshot[3]
        init_top_ask = initial_snapshot[4]
        init_tick_size = initial_snapshot[5]
    else:
        init_bid_ticks = np.array([], dtype=np.float64)
        init_bid_sizes = np.array([], dtype=np.float64)
        init_ask_ticks = np.array([], dtype=np.float64)
        init_ask_sizes = np.array([], dtype=np.float64)
        init_top_bid, init_top_ask, init_tick_size = 0, 0, 0

    snapshots = __build_orderbook_snapshots(
        dates,
        prices,
        sizes,
        is_bids,
        levels,
        tick_size_pct / 100,
        price_decimals,
        size_decimals,
        sizes_in_quoted,
        init_bid_ticks,
        init_bid_sizes,
        init_ask_ticks,
        init_ask_sizes,
        init_top_bid,
        init_top_ask,
        init_tick_size,
    )
    return snapshots


def snapshots_to_frame(snaps: list) -> pd.DataFrame:
    """
    Convert snapshots to dataframe
    """
    reindx = lambda s, d: {f"{s}{k}": v for k, v in d.items()}  # noqa: E731
    data = {
        snaps[i][0]: (
            reindx("b", dict(snaps[i][1]))
            | reindx("a", dict(snaps[i][2]))
            | {"top_bid": snaps[i][3], "top_ask": snaps[i][4], "tick_size": snaps[i][5]}
        )
        for i in range(len(snaps))
    }
    return pd.DataFrame.from_dict(data).T


def read_and_process_orderbook_updates(
    exchange: str,
    path: str,
    price_bin_pct: float,
    n_levels: int,
    sizes_in_quoted=False,
    symbols: list[str] | None = None,
    dates: slice | None = None,
    path_to_store: str | None = None,
    collect_snapshots: bool = True,
) -> dict[str, dict[datetime, pd.DataFrame]]:
    # QubxLogConfig.set_log_level("INFO")

    # - preprocess ranges
    dates_start = pd.Timestamp(dates.start if dates and dates.start else "1970-01-01")
    dates_stop = pd.Timestamp(dates.stop if dates and dates.stop else "2170-01-01")
    dates_start, dates_stop = min(dates_start, dates_stop), max(dates_start, dates_stop)

    def __process_updates_record(line: str):
        data = msgspec.json.decode(line)
        # - we need only full depth here !
        if (s_d := data.get("stream")) is not None and s_d[-6:] == "@depth":
            update = data["data"]
            if update.get("e") == "depthUpdate":
                ts = datetime.fromtimestamp(update["E"] / 1000)
                for is_bid, key in [(True, "b"), (False, "a")]:
                    for price, size in update[key]:
                        yield (ts, float(price), float(size), is_bid)

    symb_snapshots = defaultdict(dict)
    for s in Path(path).glob("*"):
        symbol = s.name.upper()

        # - skip if list is defined but symbol not in it
        if symbols and symbol not in symbols:
            continue

        instr = lookup.find_symbol(exchange.upper(), symbol)
        if not isinstance(instr, Instrument):
            logger.error(f"Instrument not found for {symbol} !")
            continue

        _latest_snapshot = None
        for d in sorted(s.glob("raw/*")):
            _d_ts = pd.Timestamp(d.name)
            if _d_ts < dates_start or _d_ts > dates_stop:
                continue

            if path_to_store and exists(_f := get_path_to_snapshots_file(path_to_store, symbol, _d_ts)):
                logger.info(f"File {_f} already exists, skipping.")
                continue

            day_updates = []
            logger.info(f"Loading {symbol} : {d.name} ... ")
            for file in sorted(d.glob("*.txt.gz")):
                try:
                    with gzip.open(file, "rt") as f:
                        try:
                            while line := f.readline():
                                for upd in __process_updates_record(line):
                                    day_updates.append(upd)
                        except Exception as exc:
                            logger.warning(f">>> Exception in processing {file.name} : {exc}")
                            # logger.opt(colors=False).error(traceback.format_exc())
                except EOFError as exc:
                    logger.error(f">>> Exception in reading {exc}")
                    logger.opt(colors=False).error(traceback.format_exc())

            if len(day_updates) == 0:
                logger.info(f"No data for {symbol} at {d.name}")
                continue

            logger.info(f"loaded {len(day_updates)} updates")

            snaps = build_orderbook_snapshots(
                day_updates,
                n_levels,
                price_bin_pct,
                instr.tick_size,
                instr.lot_size,
                sizes_in_quoted=sizes_in_quoted,
                initial_snapshot=_latest_snapshot,
            )
            _latest_snapshot = snaps[-1]

            processed_snap = snapshots_to_frame(snaps)
            t_key = pd.Timestamp(d.name).strftime("%Y-%m-%d")

            # - collect snapshots
            if collect_snapshots:
                symb_snapshots[symbol][t_key] = processed_snap

            # - save data
            if path_to_store:
                store_snapshots_to_h5(path_to_store, {symbol: {t_key: processed_snap}}, price_bin_pct, n_levels)

    return symb_snapshots


def get_combined_cumulative_snapshot(data: dict[str, dict[datetime, pd.DataFrame]], max_levs=1000000) -> pd.DataFrame:
    frms = []
    for s, dv in data.items():
        _f = {}
        for d, v in dv.items():
            ca = v.mean(axis=0).filter(regex="^a.*")[:max_levs].cumsum(axis=0)
            cb = v.mean(axis=0).filter(regex="^b.*")[::-1][:max_levs].cumsum(axis=0)
            _f[pd.Timestamp(d)] = srows(ca[::-1], cb, sort=False).to_dict()
        frms.append(pd.DataFrame.from_dict(_f, orient="index"))
    return scols(*frms, keys=data.keys())


def get_path_to_snapshots_file(path: str, symbol: str, date: str) -> str:
    _s_path = join(path, symbol.upper())
    if not os.path.exists(_s_path):
        os.makedirs(_s_path)
    return join(_s_path, pd.Timestamp(date).strftime("%Y-%m-%d")) + ".h5"


def store_snapshots_to_h5(path: str, data: dict[str, dict[str, pd.DataFrame]], p, nl):
    """
    Store orderbook data to HDF5 files
    """
    for s, v in data.items():
        for t, vd in v.items():
            logger.info(f"Storing {s} : {t}")
            vd.to_hdf(
                get_path_to_snapshots_file(path, s, t), key=f"orderbook_{str(p).replace('.', '_')}_{nl}", complevel=9
            )


def load_snapshots_from_h5(path: str, symbol: str, dates: slice | str, p: float, nl: int) -> dict[str, pd.DataFrame]:
    symbol = symbol.upper()
    if isinstance(dates, slice):
        dates_start = pd.Timestamp(dates.start if dates and dates.start else "1970-01-01")
        dates_stop = pd.Timestamp(dates.stop if dates and dates.stop else "2170-01-01")
    else:
        dates_start = pd.Timestamp(dates)
        dates_stop = pd.Timestamp(dates)
    dates_start, dates_stop = min(dates_start, dates_stop), max(dates_start, dates_stop)
    rs = {symbol: {}}
    for d in tqdm(sorted((Path(path) / symbol).glob("*.h*"))):
        _d_ts = pd.Timestamp(d.name.split(".")[0])
        if _d_ts < dates_start or _d_ts > dates_stop:
            continue
        rs[symbol][_d_ts] = pd.read_hdf(d, f"orderbook_{str(p).replace('.', '_')}_{nl}")
    return rs


def aggregate_symbol(path: str, symbol: str, p: float, nl: int, reload=False) -> pd.DataFrame | None:
    """
    Aggregate orderbook data for a symbol on a daily basis and save to HDF5 file
    """
    symbol = symbol.upper()
    result = None
    with pd.HDFStore(f"{path}/aggregated.h5", "a", complevel=9) as store:
        if reload or (f"/{symbol}" not in store.keys()):
            _f = {}
            for d in tqdm(sorted((Path(path) / symbol).glob("*.h*")), leave=False, desc=symbol):
                date = d.name.split(".")[0]
                rs = pd.read_hdf(d, f"orderbook_{str(p).replace('.', '_')}_{nl}")
                rs = rs.loc[date]
                if not rs.empty:
                    ca = rs.mean(axis=0).filter(regex="^a.*").cumsum(axis=0)
                    cb = rs.mean(axis=0).filter(regex="^b.*")[::-1].cumsum(axis=0)
                    _f[pd.Timestamp(date)] = srows(ca[::-1], cb, sort=False).to_dict()
            result = pd.DataFrame.from_dict(_f, orient="index")
            store.put(symbol, result)
    return result


def aggregate_symbols_from_list(path: str, symbols: list[str] | dict[str, Any], p: float, nl: int, reload=False):
    """
    Aggregate orderbook data for a list of symbols on a daily basis and save to HDF5 file
    """
    for s in tqdm(symbols):
        aggregate_symbol(path, s, p, nl, reload)


@njit
def accumulate_orderbook_levels(
    raw_levels: np.ndarray, buffer: np.ndarray, tick_size: float, is_bid: bool, levels: int, sizes_in_quoted: bool
) -> tuple[float, np.ndarray]:
    """
    Accumulate order book levels into price buckets based on tick size.

    Parameters:
        raw_levels (list): List of [price, size] pairs from the raw order book
        buffer (np.ndarray): Pre-allocated buffer to store accumulated sizes
        tick_size (float): The tick size to use for price bucketing
        is_bid (bool): Whether these are bid levels (True) or ask levels (False)
        levels (int): Number of price levels to include
        sizes_in_quoted (bool): Whether sizes are in quoted currency

    Returns:
        tuple: (top_price, accumulated_sizes)
    """
    if len(raw_levels) == 0:
        return 0.0, buffer

    # Find the top price (highest bid or lowest ask)
    top_price = raw_levels[0][0]

    # Calculate price buckets and accumulate sizes
    for price, size in raw_levels:
        if is_bid:
            # For bids, we floor the price to the nearest tick
            idx = int(np.floor((top_price - price) / tick_size))
        else:
            # For asks, we ceil the price to the nearest tick
            idx = int(np.floor((price - top_price) / tick_size))

        # Only accumulate if within our desired number of levels
        if 0 <= idx < levels:
            # Convert size to quoted currency if needed
            size_to_add = price * size if sizes_in_quoted else size
            buffer[idx] += size_to_add

    return top_price, buffer



import os
from pathlib import Path

import yaml
from pydantic import BaseModel, Field

from qubx.core.interfaces import IStrategy


class ConnectorConfig(BaseModel):
    connector: str
    params: dict = Field(default_factory=dict)


class ExchangeConfig(BaseModel):
    connector: str
    universe: list[str]
    params: dict = Field(default_factory=dict)
    broker: ConnectorConfig | None = None
    account: ConnectorConfig | None = None


class ReaderConfig(BaseModel):
    reader: str
    args: dict = Field(default_factory=dict)


class TypedReaderConfig(BaseModel):
    data_type: list[str] | str
    readers: list[ReaderConfig]


class RestorerConfig(BaseModel):
    type: str
    parameters: dict = Field(default_factory=dict)


class WarmupConfig(BaseModel):
    readers: list[TypedReaderConfig] = Field(default_factory=list)
    restorer: RestorerConfig | None = None


class LoggingConfig(BaseModel):
    logger: str
    position_interval: str
    portfolio_interval: str
    args: dict = Field(default_factory=dict)
    heartbeat_interval: str = "1m"


class ExporterConfig(BaseModel):
    exporter: str
    parameters: dict = Field(default_factory=dict)


class EmitterConfig(BaseModel):
    """Configuration for a single metric emitter."""

    emitter: str
    parameters: dict = Field(default_factory=dict)
    tags: dict[str, str] = Field(default_factory=dict)


class EmissionConfig(BaseModel):
    """Configuration for metric emission."""

    stats_interval: str = "1m"  # Default interval for emitting strategy stats
    stats_to_emit: list[str] | None = None  # Optional list of specific stats to emit
    emitters: list[EmitterConfig] = Field(default_factory=list)


class NotifierConfig(BaseModel):
    """Configuration for strategy lifecycle notifiers."""

    notifier: str
    parameters: dict = Field(default_factory=dict)


class HealthConfig(BaseModel):
    emit_interval: str = "10s"
    queue_monitor_interval: str = "1s"
    buffer_size: int = 5000


class LiveConfig(BaseModel):
    read_only: bool = False
    exchanges: dict[str, ExchangeConfig]
    logging: LoggingConfig
    exporters: list[ExporterConfig] | None = None
    emission: EmissionConfig | None = None
    notifiers: list[NotifierConfig] | None = None
    warmup: WarmupConfig | None = None
    health: HealthConfig = Field(default_factory=HealthConfig)


class SimulationConfig(BaseModel):
    capital: float
    instruments: list[str]
    start: str
    stop: str
    data: list[TypedReaderConfig] = Field(default_factory=list)
    commissions: dict | str | None = None
    n_jobs: int | None = None
    variate: dict = Field(default_factory=dict)
    debug: str | None = None
    run_separate_instruments: bool = False


class StrategyConfig(BaseModel):
    name: str | None = None
    description: str | list[str] | None = None
    strategy: str | list[str] | type[IStrategy]
    parameters: dict = Field(default_factory=dict)
    aux: ReaderConfig | None = None
    live: LiveConfig | None = None
    simulation: SimulationConfig | None = None


def load_strategy_config_from_yaml(path: Path | str, key: str | None = None) -> StrategyConfig:
    """
    Loads a strategy configuration from a YAML file.

    Args:
        path (str | Path): The path to the YAML file.
        key (str | None): The key to extract from the YAML file.

    Returns:
        StrategyConfig: The parsed configuration.
    """
    path = Path(os.path.expanduser(path))
    if not path.exists():
        raise FileNotFoundError(f"File {path} not found.")
    with path.open("r") as f:
        config_dict = yaml.safe_load(f)
        if key:
            config_dict = config_dict[key]
        return StrategyConfig(**config_dict)



from collections import defaultdict
from typing import Any

from qubx.core.basics import DataType, Instrument
from qubx.core.interfaces import IDataProvider, ISubscriptionManager
from qubx.utils.misc import synchronized


class SubscriptionManager(ISubscriptionManager):
    _data_providers: list[IDataProvider]
    _exchange_to_data_provider: dict[str, IDataProvider]
    _base_sub: str
    _sub_to_warmup: dict[str, str]
    _auto_subscribe: bool

    _pending_global_subscriptions: set[str]
    _pending_global_unsubscriptions: set[str]

    _pending_stream_subscriptions: dict[str, set[Instrument]]
    _pending_stream_unsubscriptions: dict[str, set[Instrument]]
    _pending_warmups: dict[tuple[str, Instrument], str]

    def __init__(
        self,
        data_providers: list[IDataProvider],
        auto_subscribe: bool = True,
        default_base_subscription: DataType = DataType.NONE,
    ) -> None:
        self._data_providers = data_providers
        self._exchange_to_data_provider = {data_provider.exchange(): data_provider for data_provider in data_providers}
        self._base_sub = default_base_subscription
        self._sub_to_warmup = {}
        self._pending_warmups = {}
        self._pending_global_subscriptions = set()
        self._pending_global_unsubscriptions = set()
        self._pending_stream_subscriptions = defaultdict(set)
        self._pending_stream_unsubscriptions = defaultdict(set)
        self._auto_subscribe = auto_subscribe

    def subscribe(self, subscription_type: str, instruments: list[Instrument] | Instrument | None = None) -> None:
        # - figure out which instruments to subscribe to (all or specific)
        if instruments is None:
            self._pending_global_subscriptions.add(subscription_type)
            return

        if isinstance(instruments, Instrument):
            instruments = [instruments]

        # - get instruments that are not already subscribed to
        _current_instruments = self.get_subscribed_instruments(subscription_type)
        instruments = list(set(instruments).difference(_current_instruments))

        # - subscribe to all existing subscriptions if subscription_type is ALL
        if subscription_type == DataType.ALL:
            subscriptions = self.get_subscriptions()
            for sub in subscriptions:
                self.subscribe(sub, instruments)
            return

        self._pending_stream_subscriptions[subscription_type].update(instruments)
        self._update_pending_warmups(subscription_type, instruments)

    def unsubscribe(self, subscription_type: str, instruments: list[Instrument] | Instrument | None = None) -> None:
        if instruments is None:
            self._pending_global_unsubscriptions.add(subscription_type)
            return

        if isinstance(instruments, Instrument):
            instruments = [instruments]

        # - subscribe to all existing subscriptions if subscription_type is ALL
        if subscription_type == DataType.ALL:
            subscriptions = self.get_subscriptions()
            for sub in subscriptions:
                self.unsubscribe(sub, instruments)
            return

        self._pending_stream_unsubscriptions[subscription_type].update(instruments)

    @synchronized
    def commit(self) -> None:
        if not self._has_operations_to_commit():
            return

        # - warm up subscriptions
        self._run_warmup()

        # - update subscriptions
        for _sub in self._get_updated_subs():
            _current_sub_instruments = set(self.get_subscribed_instruments(_sub))
            _removed_instruments = self._pending_stream_unsubscriptions.get(_sub, set())
            _added_instruments = self._pending_stream_subscriptions.get(_sub, set())

            if _sub in self._pending_global_unsubscriptions:
                _removed_instruments.update(_current_sub_instruments)

            if _sub in self._pending_global_subscriptions:
                _added_instruments.update(self.get_subscribed_instruments())

            # - subscribe collection
            _updated_instruments = _current_sub_instruments.union(_added_instruments).difference(_removed_instruments)
            _exchange_to_updated_instruments = defaultdict(set)
            _exchange_to_current_sub_instruments = defaultdict(set)
            for instr in _updated_instruments:
                _exchange_to_updated_instruments[instr.exchange].add(instr)
            for instr in _current_sub_instruments:
                _exchange_to_current_sub_instruments[instr.exchange].add(instr)

            _exchanges_to_update = set(_exchange_to_updated_instruments.keys()).union(
                _exchange_to_current_sub_instruments.keys()
            )
            for _exchange in _exchanges_to_update:
                _data_provider = self._exchange_to_data_provider[_exchange]
                _exchange_updated_instruments = _exchange_to_updated_instruments[_exchange]
                _exchange_current_instruments = _exchange_to_current_sub_instruments[_exchange]
                if _exchange_updated_instruments != _exchange_current_instruments:
                    _data_provider.subscribe(_sub, _exchange_updated_instruments, reset=True)

            # - unsubscribe instruments
            _exchange_to_removed_instruments = defaultdict(set)
            for instr in _removed_instruments:
                _exchange_to_removed_instruments[instr.exchange].add(instr)

            for _exchange, _exchange_removed_instruments in _exchange_to_removed_instruments.items():
                _data_provider = self._exchange_to_data_provider[_exchange]
                _data_provider.unsubscribe(_sub, _exchange_removed_instruments)

        # - clean up pending subs and unsubs
        self._pending_stream_subscriptions.clear()
        self._pending_stream_unsubscriptions.clear()
        self._pending_global_subscriptions.clear()
        self._pending_global_unsubscriptions.clear()

    def has_subscription(self, instrument: Instrument, subscription_type: str) -> bool:
        _data_provider = self._exchange_to_data_provider[instrument.exchange]
        return _data_provider.has_subscription(instrument, subscription_type)

    def get_subscriptions(self, instrument: Instrument | None = None) -> list[str]:
        _data_provider = (
            self._exchange_to_data_provider[instrument.exchange] if instrument is not None else self._data_providers[0]
        )
        return list(
            set(_data_provider.get_subscriptions(instrument))
            | {self.get_base_subscription()}
            | self._pending_global_subscriptions
        )

    def get_subscribed_instruments(self, subscription_type: str | None = None) -> list[Instrument]:
        _current_instruments = []
        for _data_provider in self._data_providers:
            _current_instruments.extend(_data_provider.get_subscribed_instruments(subscription_type))
        return _current_instruments

    def get_base_subscription(self) -> str:
        return self._base_sub

    def set_base_subscription(self, subscription_type: str) -> None:
        self._base_sub = subscription_type

    def get_warmup(self, subscription_type: str) -> str | None:
        return self._sub_to_warmup.get(subscription_type)

    def set_warmup(self, configs: dict[Any, str]) -> None:
        for subscription_type, period in configs.items():
            self._sub_to_warmup[subscription_type] = period

    @property
    def auto_subscribe(self) -> bool:
        return self._auto_subscribe

    @auto_subscribe.setter
    def auto_subscribe(self, value: bool) -> None:
        self._auto_subscribe = value

    def _get_updated_subs(self) -> list[str]:
        return list(
            set(self._pending_stream_unsubscriptions.keys())
            | set(self._pending_stream_subscriptions.keys())
            | self._pending_global_subscriptions
            | self._pending_global_unsubscriptions
        )

    def _has_operations_to_commit(self) -> bool:
        return any(
            (
                self._pending_stream_unsubscriptions,
                self._pending_stream_subscriptions,
                self._pending_global_subscriptions,
                self._pending_global_unsubscriptions,
            )
        )

    def _update_pending_warmups(self, subscription_type: str, instruments: list[Instrument]) -> None:
        # TODO: refactor pending warmups in a way that would allow to subscribe and then call set_warmup in the same iteration
        # - ohlc is handled separately
        if DataType.from_str(subscription_type) != DataType.OHLC:
            _warmup_period = self._sub_to_warmup.get(subscription_type)
            if _warmup_period is not None:
                for instrument in instruments:
                    self._pending_warmups[(subscription_type, instrument)] = _warmup_period

        # - if base subscription, then we need to fetch historical OHLC data for warmup
        if subscription_type == self._base_sub:
            self._pending_warmups.update(
                {
                    (sub, instrument): period
                    for sub, period in self._sub_to_warmup.items()
                    for instrument in instruments
                    if DataType.OHLC == sub
                }
            )

    def _run_warmup(self) -> None:
        # - handle warmup for global subscriptions
        for _data_provider in self._data_providers:
            _subscribed_instruments = set(_data_provider.get_subscribed_instruments())
            _new_instruments = (
                set.union(*self._pending_stream_subscriptions.values()) if self._pending_stream_subscriptions else set()
            )

            for sub in self._pending_global_subscriptions:
                _warmup_period = self._sub_to_warmup.get(sub)
                if _warmup_period is None:
                    continue
                _sub_instruments = _data_provider.get_subscribed_instruments(sub)
                _add_instruments = _subscribed_instruments.union(_new_instruments).difference(_sub_instruments)
                for instr in _add_instruments:
                    self._pending_warmups[(sub, instr)] = _warmup_period

            # TODO: think about appropriate handling of timeouts
            _data_provider.warmup(self._pending_warmups.copy())

        self._pending_warmups.clear()



from qubx.backtester.ome import SimulatedExecutionReport
from qubx.backtester.simulated_exchange import ISimulatedExchange
from qubx.core.basics import (
    CtrlChannel,
    Instrument,
    Order,
)
from qubx.core.interfaces import IBroker

from .account import SimulatedAccountProcessor


class SimulatedBroker(IBroker):
    channel: CtrlChannel

    _account: SimulatedAccountProcessor
    _exchange: ISimulatedExchange

    def __init__(
        self,
        channel: CtrlChannel,
        account: SimulatedAccountProcessor,
        simulated_exchange: ISimulatedExchange,
    ) -> None:
        self.channel = channel
        self._account = account
        self._exchange = simulated_exchange

    @property
    def is_simulated_trading(self) -> bool:
        return True

    def send_order(
        self,
        instrument: Instrument,
        order_side: str,
        order_type: str,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        time_in_force: str = "gtc",
        **options,
    ) -> Order:
        # - place order at exchange and send exec report to data channel
        self._send_execution_report(
            report := self._exchange.place_order(
                instrument, order_side, order_type, amount, price, client_id, time_in_force, **options
            )
        )
        return report.order

    def send_order_async(
        self,
        instrument: Instrument,
        order_side: str,
        order_type: str,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        time_in_force: str = "gtc",
        **optional,
    ) -> None:
        self.send_order(instrument, order_side, order_type, amount, price, client_id, time_in_force, **optional)

    def cancel_order(self, order_id: str) -> Order | None:
        self._send_execution_report(order_update := self._exchange.cancel_order(order_id))
        return order_update.order if order_update is not None else None

    def cancel_orders(self, instrument: Instrument) -> None:
        raise NotImplementedError("Not implemented yet")

    def update_order(self, order_id: str, price: float | None = None, amount: float | None = None) -> Order:
        raise NotImplementedError("Not implemented yet")

    def _send_execution_report(self, report: SimulatedExecutionReport | None):
        if report is None:
            return

        self.channel.send((report.instrument, "order", report.order, False))
        if report.exec is not None:
            self.channel.send((report.instrument, "deals", [report.exec], False))

    def exchange(self) -> str:
        return self._exchange.exchange_id
